"""
Quantiphyse - PCA reduction process

Copyright (c) 2013-2020 University of Oxford

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import six

from quantiphyse.data.extras import MatrixExtra
from quantiphyse.utils import QpException
from quantiphyse.processes import Process

import numpy as np

class HistogramProcess(Process):
    """
    Calculate histogram for a data set 
    """
    
    PROCESS_NAME = "Histogram"
    
    def __init__(self, ivm, **kwargs):
        Process.__init__(self, ivm, **kwargs)

    def run(self, options):
        data_items = options.pop('data', None)
        if data_items is None:
            data_items = self.ivm.data.keys()
        elif isinstance(data_items, six.string_types):
            data_items = [data_items,]

        if not data_items:
            raise QpException("No data to calculate histogram")
        
        data_items = [self.ivm.data[name] for name in data_items]
        roi = self.get_roi(options, use_current=False)
        
        sel_region = options.pop('region', None)
        dmin = options.pop('min', None)
        dmax = options.pop('max', None)
        bins = options.pop('bins', 100)
        vol = options.pop('vol', None)
        prob = options.pop('yscale', 'count').lower().startswith("prob")
        output_name = options.pop('output-name', "histogram")

        yvals, col_headers = {}, ["left", "right", "centre",]
        for data in data_items:
            if vol is not None:
                rawdata = data.volume(vol)
            else:
                rawdata = data.raw()
            hrange = [dmin, dmax]
            if dmin is None: hrange[0] = np.min(rawdata)
            if dmax is None: hrange[1] = np.max(rawdata)

            if roi is None:
                roi_fordata = np.ones(data.grid.shape)
                regions = {1 : ""}
            else:
                roi_fordata = roi.resample(data.grid).raw()
                regions = roi.regions

            for region, region_name in regions.items():
                if sel_region is not None and region != sel_region:
                    self.debug("Ignoring region %i", region)
                    continue

                region_data = rawdata[roi_fordata == region]
                data_vals, edges = np.histogram(region_data, bins=bins, range=hrange, density=prob)
                    
                if region_name:
                    name = "%s\n%s" % (data.name, region_name)
                else:
                    name = data.name

                col_headers.append(name)
                yvals[name] = data_vals
            
        xvals = [[edges[idx], edges[idx+1], (edges[idx]+edges[idx+1])/2] for idx in range(len(edges)-1)]
        rows = []
        for idx, row in enumerate(xvals):
            for name in col_headers[3:]:
                row += [yvals[name][idx],]
            rows.append(row)

        if not options.pop('no-extras', False):
            self.debug("Adding %s" % output_name)
            extra = MatrixExtra(output_name, rows, col_headers=col_headers)
            self.debug(str(extra))
            self.ivm.add_extra(output_name, extra)



"""
Quantiphyse - PCA implementation

Copyright (c) 2013-2020 University of Oxford

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from __future__ import division, print_function, absolute_import

import numpy as np
from sklearn.decomposition import PCA
from scipy.ndimage.filters import gaussian_filter1d

from quantiphyse.utils import QpException, LogSource
from . import normalisation as norm

def _flat_data_to_image(data, roi):
    # return features in image format
    image_shape = list(roi.shape)
    if data.ndim == 2:
        image_shape.append(data.shape[1])
    image = np.zeros(image_shape)
    image[roi] = data
    return image

class PcaFeatReduce(LogSource):
    """
    Extract PCA features from 4D image data

    Thin wrapper around sklearn.decomposition.PCA
    """

    def __init__(self, n_components, norm_modes=True, norm_input=False, norm_type='perc'):
        """
        :param data: 4D image to be processed
        :param roi: Optional sub region to extract
        """
        LogSource.__init__(self)

        # Variables
        self.pca = PCA(n_components=n_components)
        self.norm_modes = norm_modes
        self.norm_input = norm_input
        self.norm_type = norm_type

    def get_training_features(self, data, roi=None, smooth_timeseries=None, feature_volume=False):
        """
        Train PCA reduction on data and return features for each voxel

        :param data: 4D data set
        :param roi: Optional 3D ROI
        :param smooth_timeseries: Optional sigma for 1D Gaussian smoothing of each voxel timeseries
        :param feature_volume: determines whether the features are returned as a list or an image

        :return: If ``feature_volume``, 4D array with the same 3d dimensions as data and 
                 4th dimension=number of PCA components.
                 Otherwise, 2D array whose first dimension is unmasked voxels and 2nd dimension
                 is the PCA components
        """
        data_inmask, roi = self._mask(data, roi, smooth_timeseries)

        self.debug("Using PCA dimensionality reduction")
        reduced_data = self.pca.fit_transform(data_inmask)
        self.debug("Number of components", reduced_data.shape[1])

        if self.norm_modes:
            self.debug("Normalising PCA modes between 0 and 1")
            reduced_data = norm.normalise(reduced_data, "indiv")

        if not feature_volume:
            return reduced_data

        return _flat_data_to_image(reduced_data, roi)

    def get_projected_test_features(self, data, roi=None, smooth_timeseries=None, feature_volume=False):
        """
        Return features for each voxel from previously trained PCA modes

        :param data: 4D data set
        :param roi: Optional 3D ROI
        :param feature_volume: determines whether the features are returned as a list or an image

        :return: If ``feature_volume``, 4D array with the same 3d dimensions as data and 
                 4th dimension=number of PCA components.
                 Otherwise, 2D array whose first dimension is unmasked voxels and 2nd dimension
                 is the PCA components
        """
        data_inmask, roi = self._mask(data, roi, smooth_timeseries)
        
        #Projecting the data using training set PCA
        if data_inmask.shape[1] != self.pca.mean_.shape[0]:
            raise QpException("Input data length does not match previous training data")
            
        reduced_data = self.pca.transform(data_inmask)

        # Scaling features
        if self.norm_modes:
            self.debug("Normalising PCA modes")
            reduced_data = norm.normalise(reduced_data, "indiv")

        if not feature_volume:
            return reduced_data
        
        return _flat_data_to_image(reduced_data, roi)

    def explained_variance(self, cumulative=False):
        """
        Return the variance explained by including each of the modes cumulatively
        
        :param cumulative: If True, variance explained will be cumulative (i.e.
                           variance explained by all modes up to and including each
                           mode)
        :return: Sequence of numbers giving variance explained for each component. 
        """ 
        if cumulative:
            return [np.sum(self.pca.explained_variance_ratio_[:n+1]) 
                    for n in range(self.pca.n_components_)]
        else:
            return self.pca.explained_variance_ratio_

    def modes(self):
        """
        :return: Fitted PCA modes
        """
        #return norm.normalise(self.pca.components_, "indiv")
        return [comp + self.pca.mean_ for comp in self.pca.components_]

    def mean(self):
        """
        :return: Training data mean
        """
        #return np.squeeze(norm.normalise(np.expand_dims(self.pca.mean_, axis=0), "indiv"))
        return self.pca.mean_

    def _mask(self, data, roi, smooth_timeseries):
        if roi is None:
            roi = np.ones(data.shape[0:-1], dtype=bool)
        else:
            roi = np.array(roi, dtype=np.bool)
        data_inmask = data[roi]

        if self.norm_input:
            data_inmask = norm.normalise(data_inmask, self.norm_type)

        if smooth_timeseries is not None:
            data_inmask = gaussian_filter1d(data_inmask, sigma=smooth_timeseries, axis=-1)
        return data_inmask, roi



"""
Quantiphyse - Functions for querying and reading locally installed files

Copyright (c) 2013-2020 University of Oxford

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import sys
import os
import logging
import warnings

LOCAL_FILE_PATH = ""
DEFAULT_SAVE_DIR = ""
LOG = logging.getLogger(__name__)

def set_local_file_path():
    """
    Initialize the file path to use when looking for local files, e.g. icons, plugins, etc

    This depends on whether we are in a frozen executable or running from a script so 
    various possibilities
    """
    global LOCAL_FILE_PATH
    LOCAL_FILE_PATH = ""
    if hasattr(sys, 'frozen'):
        # File is frozen (packaged apps)
        LOG.debug("Frozen executable")
        if hasattr(sys, '_MEIPASS'):
            LOCAL_FILE_PATH = sys._MEIPASS
        elif hasattr(sys, '_MEIPASS2'):
            LOCAL_FILE_PATH = sys._MEIPASS2
        elif sys.frozen == 'macosx_app':
            LOCAL_FILE_PATH = os.getcwd() + '/quantiphyse'
        else:
            LOCAL_FILE_PATH = os.path.dirname(sys.executable)
    else:
        # Running from a script
        LOCAL_FILE_PATH = os.path.join(os.path.dirname(__file__), os.pardir)
        
    if LOCAL_FILE_PATH == "":
        # Use local working directory otherwise
        LOG.warn("Reverting to current directory as local path")
        LOCAL_FILE_PATH = os.getcwd()

    LOG.debug("Local directory: %s", LOCAL_FILE_PATH)

def local_file_path():
    """
    :return: The path to locally installed files
    """
    return LOCAL_FILE_PATH

def get_local_file(name, loc=None):
    """
    Get path to a file relative to the main Quantiphyse folder

    If location is not None, use it to determine the local root folder 
    (e.g. use __file__ to get file local to a python plugin module)
    """
    if loc is None:
        loc = LOCAL_FILE_PATH
    else:
        loc = os.path.dirname(loc)
    return os.path.abspath(os.path.join(loc, name))

def local_file_from_drop_url(url):
    """
    Get the local file path associated with a drag/drop URL

    This is platform-dependent so put into it's own function

    :return: Local file path
    """
    if sys.platform.startswith("darwin"):
        # OSx specific changes to allow drag and drop
        from Cocoa import NSURL
        url = str(url.toString()).replace(" ", "%20")
        url_obj = NSURL.URLWithString_(url)
        if url_obj:
            return str(url_obj.filePathURL().path())
        else:
            LOG.warn("Failed to convert local file URL: %s", url)
            return None
    else:
        path = str(url.toLocalFile())
        if sys.platform.startswith("win"):
            # QT bug with UNC paths. We can't fix this generally but we want
            # the WSL specific case to work, so detect that at least
            wslpath = "\\\\wsl$\\" + path
            if not os.path.exists(path) and os.path.exists(wslpath):
                return wslpath
        return path

def get_icon(name, icon_dir=None):
    """
    Get path to the named icon
    """
    name, extension = os.path.splitext(name)
    if extension == "":
        extension = ".png"
    tries = []
    if icon_dir is not None: 
        tries.append(os.path.join(icon_dir, "%s%s" % (name, extension)))
    tries.append(os.path.join(LOCAL_FILE_PATH, "icons", "%s%s" % (name, extension)))
    for icon_file in tries:
        if os.path.isfile(icon_file): return icon_file

def get_lib_fname(name):
    """ Get file name for named shared library on current platform """
    if sys.platform.startswith("win"):
        return "%s.dll" % name
    elif sys.platform.startswith("darwin"):
        return "lib%s.dylib" % name
    else:
        return "lib%s.so" % name

def get_local_shlib(name, loc):
    """
    Get a named shared library which is stored locally to another file
    """
    return get_local_file(get_lib_fname(name), loc)
    
def default_save_dir():
    return DEFAULT_SAVE_DIR

def set_default_save_dir(save_dir):
    global DEFAULT_SAVE_DIR
    DEFAULT_SAVE_DIR = save_dir


"""
Quantiphyse - Subclass of QpData for handling Nifti data

Copyright (c) 2013-2020 University of Oxford

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
from __future__ import division, print_function

import os
import logging
import traceback

import nibabel as nib
import numpy as np

from .qpdata import DataGrid, QpData, NumpyData, Metadata

LOG = logging.getLogger(__name__)

QP_NIFTI_EXTENSION_CODE = 42

class NiftiData(QpData):
    """
    QpData from a Nifti file
    """
    def __init__(self, fname):
        nii = nib.load(fname)
        shape = list(nii.shape)
        while len(shape) < 3:
            shape.append(1)

        if len(shape) > 3:
            nvols = shape[3]
        else:
            nvols = 1

        self.rawdata = None
        self.voldata = None
        self.nifti_header = nii.header
        metadata = None
        for ext in self.nifti_header.extensions:
            if ext.get_code() == QP_NIFTI_EXTENSION_CODE:
                import yaml
                LOG.debug("Found QP metadata: %s", ext.get_content())
                try:
                    yaml_data = yaml.load(ext.get_content())
                    LOG.debug("YAML data: %s", yaml_data)
                    if isinstance(yaml_data, Metadata):
                        # Old style
                        metadata = yaml_data
                    else:
                        # New style
                        metadata = yaml_data[0]["QpMetadata"]
                    LOG.debug(metadata)
                except:
                    # Never fail to load a file just because of metadata
                    LOG.warn("Failed to read Quantiphyse metadata")
                    LOG.warn(ext.get_content())
                    traceback.print_exc()

        xyz_units, vol_units = "mm", None
        units = nii.header.get_xyzt_units()
        if units:
            xyz_units = units[0]
            if len(units) > 1:
                vol_units = units[1]

        vol_scale = 1.0
        zooms = nii.header.get_zooms()
        if zooms and len(zooms) > 3:
            vol_scale = zooms[3]

        grid = DataGrid(shape[:3], nii.header.get_best_affine(), units=xyz_units)
        QpData.__init__(self, fname, grid, nvols, vol_unit=vol_units, vol_scale=vol_scale, fname=fname, metadata=metadata)

    def raw(self):
        # NB: copy() converts data to an in-memory array instead of a numpy file memmap.
        # Appears to improve speed drastically as well as stop a bug with accessing the subset of the array
        # memmap has been designed to save space on ram by keeping the array on the disk but does
        # horrible things with performance, and analysis especially when the data is on the network.
        if self.rawdata is None:
            nii = nib.load(self.fname)
            #self.rawdata = nii.get_data().copy()
            self.rawdata = nii.get_data()
            self.rawdata = self._correct_dims(self.rawdata)

        self.voldata = None
        return self.rawdata
        
    def volume(self, vol, qpdata=False):
        vol = min(vol, self.nvols-1)
        if self.nvols == 1:
            ret = self.raw()
        elif self.rawdata is not None:
            ret = self.rawdata[:, :, :, vol]
        else:
            if self.voldata is None:
                self.voldata = [None,] * self.nvols
            if self.voldata[vol] is None:
                nii = nib.load(self.fname)
                self.voldata[vol] = self._correct_dims(nii.dataobj[..., vol])
            ret = self.voldata[vol]

        if qpdata:
            return NumpyData(ret, grid=self.grid, name="%s_vol_%i" % (self.name, vol))
        else:
            return ret

    def _correct_dims(self, arr):
        while arr.ndim < 3:
            arr = np.expand_dims(arr, -1)

        if self.metadata.get("raw_2dt", False) and arr.ndim == 3:
            # Single-slice, interpret 3rd dimension as time
            arr = np.expand_dims(arr, 2)

        if arr.ndim == 4 and arr.shape[3] == 1:
            arr = np.squeeze(arr, axis=-1)
        return arr

def save(data, fname, grid=None, outdir=""):
    """
    Save data to a file
    
    :param data: QpData instance
    :param fname: File name
    :param grid: If specified, grid to save the data on
    :param outdir: Optional output directory if fname is not absolute
    """
    if grid is None:
        grid = data.grid
        arr = data.raw().copy()
    else:
        arr = data.resample(grid).raw().copy()
        
    if hasattr(data, "nifti_header"):
        header = data.nifti_header.copy()
    else:
        header = None

    img = nib.Nifti1Image(arr, grid.affine, header=header)
    img.update_header()
    if data.metadata:
        from quantiphyse.utils.batch import to_yaml
        yaml_metadata = to_yaml({"QpMetadata" : data.metadata})
        LOG.debug("Writing metadata: %s", yaml_metadata)
        extensions = nib.nifti1.Nifti1Extensions([ext for ext in img.header.extensions if ext.get_code() != QP_NIFTI_EXTENSION_CODE])
        extensions.append(nib.nifti1.Nifti1Extension(QP_NIFTI_EXTENSION_CODE, yaml_metadata.encode('utf-8')))
        img.header.extensions = extensions

    if not fname:
        fname = data.name
        
    _, extension = os.path.splitext(fname)
    if extension == "":
        fname += ".nii"
        
    if not os.path.isabs(fname):
        fname = os.path.join(outdir, fname)

    dirname = os.path.dirname(fname)
    if not os.path.exists(dirname):
        os.makedirs(dirname)

    LOG.debug("Saving %s as %s", data.name, fname)
    img.to_filename(fname)
    data.fname = fname



"""
Quantiphyse - Exception subclasses

Copyright (c) 2013-2020 University of Oxford

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

class QpException(Exception):
    """ 
    Base class for quantiphyse-specific exceptions

    These are, by definition, exceptions which are expected under certain
    input conditions. They do not immediately imply a defect in the code
    (although a defect in the code might prompt one of these exceptions
    to be thrown). For example, invalid user input, malformed input files,
    etc, might all throw a QpException. No stack trace will be displayed
    for these exceptions, and when throwing the message/detail should
    be written so it can be understood by an end-user.

    By contrast, any exception within the normal Python exception hierarchy
    implies a bug in the code (even if the bug consists of 'need to throw a
    QpException instead). A stack trace will be displayed.
    
    Currently we do not define more specific subclasses although this
    might be useful in the future
    """

    def __init__(self, msg, detail=None):
        super(QpException, self).__init__(msg)
        self.msg = msg
        self.detail = detail

    def __str__(self):
        """ 
        String representation of exception
        
        Don't want QpException: in front of message as this type of exception
        is for the user to see
        """
        return self.msg



"""
Quantiphyse - Basic classes for Extras

Copyright (c) 2013-2020 University of Oxford

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import csv
import six

class Extra(object):
    """
    Base class for things which can be stored in the IVM apart from data sets.

    Essentially the only thing an Extra needs to be able to do is be written
    out as a string. We force subclasses to override this.

    We also provide a metadata dictionary - ideally extras should write their
    metadata in __str__ but in practice this may not be possible when we want
    the output to be compatible with external programs (e.g. writing out a
    matrix as TSV)

    In the future we might expand the Extra base class to define other behaviours,
    e.g. flexible saving to an output file, alternative capabilities... But we
    want to keep things simple for now while we figure out what use can
    be made of them.

    Currently the main uses for Extras are:

      - Tabular output, e.g. data statistics which we might want to write out to a file
      - Matrix outputs, e.g. affine transformations which are the output of a registration
    """
    def __init__(self, name):
        self.name = name
        self.metadata = {}

    def __str__(self):
        raise NotImplementedError("Subclasses of Extra must implement __str__")

class NumberListExtra(Extra):
    """
    Extra which represents a list of numbers
    """
    def __init__(self, name, values):
        """
        :param name: Extra name
        :param values: Sequence of numeric values
        """
        Extra.__init__(self, name)

        # Check all values are numeric
        [float(v) for v in values]
        self.values = values

    def __str__(self):
        """
        Output as simple space-separated list
        """
        return " ".join([str(v) for v in self.values])

class MatrixExtra(Extra):
    """
    Extra which represents a 2D matrix with optional row and column headers
    """
    def __init__(self, name, arr, row_headers=(), col_headers=()):
        """
        :param name: Extra name
        :param arr: List-of-lists or 2D Numpy array containing matrix data
        :param row_headers: Optional sequence of row headers
        :param col_headers: Optional sequence of column headers
        """
        Extra.__init__(self, name)
        if len(arr) == 0:
            raise ValueError("No matrix data given")
        if row_headers and len(row_headers) != len(arr):
            raise ValueError("Incorrect number of row headers given")
        if col_headers and len(col_headers) != len(arr[0]):
            raise ValueError("Incorrect number of column headers given")

        self.arr = arr
        self.row_headers = list(row_headers)
        self.col_headers = list(col_headers)

    def __str__(self):
        """
        Convert matrix to a string in TSV format
        """
        stream = six.StringIO()
        writer = csv.writer(stream, delimiter='\t', lineterminator='\n')
        if self.col_headers:
            if self.row_headers:
                writer.writerow([" "] + self.col_headers)
            else:
                writer.writerow(self.col_headers)

        for row in self.arr:
            if self.row_headers:
                writer.writerow([" "] + list(row))
            else:
                writer.writerow(list(row))

        return stream.getvalue()

class DataFrameExtra(Extra):
    """
    Extra which represents a Pandas data frame

    This is useful for representing general tabular data.
    """
    def __init__(self, name, df):
        """
        :param name: Extra name
        :param arr: List-of-lists or 2D Numpy array containing matrix data
        :param row_headers: Optional sequence of row headers
        :param col_headers: Optional sequence of column headers
        """
        Extra.__init__(self, name)
        self.df = df

    def __str__(self):
        stream = six.StringIO()
        self.df.to_csv(stream, sep='\t')
        return stream.getvalue()



"""
Quantiphyse - Rag-bag of utility functions - should be moved into specific modules

Copyright (c) 2013-2020 University of Oxford

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from __future__ import division, print_function

import sys
import tempfile
import logging

from matplotlib import cm
import numpy as np
import pandas as pd

from PySide2 import QtGui, QtCore, QtWidgets

from quantiphyse.data.extras import DataFrameExtra
from .exceptions import QpException
from .logger import LogSource
from .plugins import get_plugins
from .local import set_default_save_dir, default_save_dir, get_local_file, get_local_shlib, get_icon, set_local_file_path, local_file_from_drop_url

__all__ = ["QpException", "LogSource", "get_plugins", "get_local_file", "get_local_shlib", "get_icon", "set_local_file_path", "local_file_from_drop_url",]

DEFAULT_SIG_FIG = 4
LOG = logging.getLogger(__name__)

def norecurse(fn):
    """
    Decorator which prevents recursive calling of an object method
    """
    def _wrapper(*args):
        self = args[0]
        if not self.updating:
            try:
                self.updating = True
                return fn(*args)
            finally:
                self.updating = False
    return _wrapper

def ifnone(obj, alt):
    """
    Convenience function to return an alternative if an object is None

    Why isn't this in the standard library!

    :param obj: Object
    :param alt: Alternative
    :return: obj if not None, otherwise alt
    """
    if obj is None:
        return alt
    else:
        return obj

def show_help(section="", base='http://quantiphyse.readthedocs.io/en/latest/'):
    """
    Open the help page at a specified page

    :param section: Section ID, if not specified open main index page
    :param base: Base URL for documentation, may be version dependent
    """ 
    if section != "" and not section.endswith(".html") and not section.endswith("/"): 
        section += ".html"
    link = base + section
    QtGui.QDesktopServices.openUrl(QtCore.QUrl(link, QtCore.QUrl.TolerantMode))

def load_matrix(filename):
    """
    Load a file containing a matrix of numbers in ASCII text

    :return: Matrix of numbers (list of list, not Numpy array), number of rows, number of columns
    """
    try:
        try:
            a = np.loadtxt(filename)
        except ValueError:
            a = np.loadtxt(filename, delimiter=",")
    except ValueError as exc:
        raise QpException("Invalid matrix data: %s" % str(exc))

    a = np.atleast_2d(a)
    if a.ndim != 2:
        raise QpException("Matrix data was not 1D or 2D")

    return a, a.shape[0], a.shape[1]

def text_to_matrix(text):
    """
    Convert ASCII text into a matrix of numbers
    
    :return: Matrix of numbers (list of list, not Numpy array), number of rows, number of columns
    """
    fvals = []
    ncols = -1
    lines = text.splitlines()
    for line in lines:
        # Discard comments
        line = line.split("#", 1)[0].strip()
        # Split by commas or spaces
        vals = line.replace(",", " ").split()
        # Ignore empty lines
        if not vals: continue
        # Check correct number of columns
        if ncols < 0: ncols = len(vals)
        elif len(vals) != ncols:
            raise QpException("File must contain a matrix of numbers with fixed size (rows/columns)")
        # Check all data is numeric
        for val in vals:
            try:
                float(val)
            except:
                raise QpException("Non-numeric value '%s' found in matrix" % val)
        fvals.append([float(v) for v in vals])     
    
    return fvals, len(fvals), ncols

def write_temp_matrix(prefix, matrix):
    """
    Write a matrix to a temporary file, returning the filename
    """
    temp_file = tempfile.NamedTemporaryFile(prefix=prefix, delete=False)
    LOG.debug("matrix: %s", prefix)
    for row in matrix:
        line = " ".join([str(v) for v in row]) + "\n"
        LOG.debug(line)
        temp_file.write(line)
    temp_file.close()
    return temp_file.name

def get_version():
    """
    :return: Quantiphyse version number
    """
    try:
        from .._version import __version__
        return __version__
    except ImportError:
        LOG.warn(sys.exc_info()[1])
        return "<version not found>"

def table_to_extra(tabmod, name):
    """ 
    Turn a QT table model into a DataFrameExtra
    """
    cols = range(tabmod.columnCount())
    rows = range(tabmod.rowCount())
    columns = [tabmod.horizontalHeaderItem(col).text().replace("\n", " ") for col in cols]
    index = [tabmod.verticalHeaderItem(row).text() for row in rows] 

    rowdata = []
    for row in rows:
        rowdata.append([tabmod.item(row, col).text() for col in cols])

    df = pd.DataFrame(rowdata, index=index, columns=columns)
    return DataFrameExtra(name, df)

def copy_table(tabmod):
    """ Copy a QT table model to the clipboard in a form suitable for paste into Excel etc """
    clipboard = QtWidgets.QApplication.clipboard()
    tsv = str(table_to_extra(tabmod, ""))
    clipboard.setText(tsv)

def sf(num, sig_fig=None):
    """ Format a number as a string to a given number of sig figs """
    if sig_fig is None: 
        global DEFAULT_SIG_FIG
        sig_fig = DEFAULT_SIG_FIG
    fmt = "%%.%ig" % sig_fig
    return fmt % float(num)

def remove_nans(data, fillvalue=0):
    """
    Check for and remove nans from data arrays
    """
    notnans = np.isfinite(data)
    if not np.all(notnans):
        data[np.logical_not(notnans)] = fillvalue



"""
Quantiphyse - Widgets for PCA reduction

Copyright (c) 2013-2020 University of Oxford

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import numpy as np
from PySide2 import QtGui, QtCore, QtWidgets

from quantiphyse.gui.widgets import QpWidget, TitleWidget, RunWidget
from quantiphyse.gui.options import OptionBox, DataOption, NumericOption, OutputNameOption
from quantiphyse.gui.plot import Plot
from quantiphyse.utils import sf

class PcaWidget(QpWidget):
    """
    PCA widget
    """
    def __init__(self, **kwargs):
        super(PcaWidget, self).__init__(name="PCA", icon="pca",
                                        desc="PCA reduction", group="Processing", **kwargs)
        
    def init_ui(self):
        vbox = QtWidgets.QVBoxLayout()
        self.setLayout(vbox)

        title = TitleWidget(self, title="PCA reduction", subtitle="Principal Component Analysis for 4D data")
        vbox.addWidget(title)

        self._options = OptionBox("Options")
        self._options.add("Data", DataOption(self.ivm, include_3d=False), key="data")
        self._options.add("ROI", DataOption(self.ivm, data=False, rois=True), key="roi")
        self._options.add("Number of components", NumericOption(minval=1, intonly=True, default=4), key="n-components")
        self._options.add("Output name", OutputNameOption(src_data=self._options.option("data"), suffix="_pca"), key="output-name")
        self._options.option("data").sig_changed.connect(self._data_changed)
        vbox.addWidget(self._options)

        self._run = RunWidget(self)
        self._run.sig_postrun.connect(self._postrun)
        vbox.addWidget(self._run)

        self.plot = Plot(qpo=None, parent=self, title="PCA modes")

        self.variance_model = QtGui.QStandardItemModel()
        variance_table = QtWidgets.QTableView()
        variance_table.verticalHeader().hide()
        variance_table.setModel(self.variance_model)

        tabs = QtWidgets.QTabWidget()
        tabs.addTab(self.plot, "PCA modes")
        tabs.addTab(variance_table, "Explained variance")
        tabs.setCurrentWidget(self.plot)
        vbox.addWidget(tabs)

        vbox.addStretch(1)
        self._data_changed()  
    
    def processes(self):
        return {"PCA" : self._options.values()}
        
    def _data_changed(self):
        self._run.setEnabled(self._options.option("data").value in self.ivm.data)

    def _postrun(self):
        self._update_plot()
        self._update_table()

    def _update_plot(self):
        self.plot.clear()
        extra = self.ivm.extras.get(self._options.option("output-name").value + "_modes", None)
        if extra is not None:
            arr = np.array(extra.arr)
            for idx in range(arr.shape[1]-1):
                self.plot.add_line(arr[:, idx], name="Mode %i" % idx)
            self.plot.add_line(arr[:, -1], name="Mean", line_col=(255, 0, 0), line_width=3.0)

    def _update_table(self):
        self.variance_model.clear()
        extra = self.ivm.extras.get(self._options.option("output-name").value + "_variance", None)
        if extra is not None:
            self.debug(str(extra))
            for idx, header in enumerate(extra.col_headers):
                self.variance_model.setHorizontalHeaderItem(idx, QtGui.QStandardItem(header))
           
            for idx, variance in enumerate(extra.arr):
                self.variance_model.setItem(idx, 0, QtGui.QStandardItem(str(variance[0])))
                self.variance_model.setItem(idx, 1, QtGui.QStandardItem(sf(variance[1])))
                self.variance_model.setItem(idx, 2, QtGui.QStandardItem(sf(variance[2])))



"""
Quantiphyse - ROI builder tools

Copyright (c) 2013-2020 University of Oxford

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from __future__ import division, unicode_literals, absolute_import, print_function

import numpy as np
import skimage.segmentation
import scipy.ndimage

from PySide2 import QtGui, QtCore, QtWidgets

from quantiphyse.data import NumpyData
from quantiphyse.gui.widgets import OverlayCombo, RoiCombo, NumericOption, NumericSlider
from quantiphyse.gui.viewer.pickers import PickMode
from quantiphyse.utils import LogSource
from quantiphyse.processes.feat_pca import PcaFeatReduce

class Tool(LogSource):
    """
    An ROI builder tool
    """

    def __init__(self, name, tooltip=""):
        """
        :param name: Name of tool
        :param tooltip: Descriptive text
        """
        LogSource.__init__(self)
        self.name = name
        self.tooltip = tooltip

    def selected(self):
        """
        Called when tool is selected by the user 
        """
        self.ivl.set_picker(PickMode.SINGLE)

    def deselected(self):
        """
        Called when tool is deselected by the user 

        *Must* reliably disconnect any signals connected in the selected() method
        """
        self.ivl.set_picker(PickMode.SINGLE)

    def interface(self):
        """
        Return a user interface 

        :return: QtCore.QLayout suitable for embedding in toolbox
        """
        grid = QtWidgets.QGridLayout()
        grid.addWidget(QtWidgets.QLabel(self.tooltip), 0, 0, 1, 2)
        return grid

    def _show_builder_roi(self):
        if self.builder.roiname in self.ivm.rois:
            self.ivm.set_current_roi(self.builder.roiname)

class CrosshairsTool(Tool):
    """
    'Tool' which does nothing but allow the user to navigate the views without
    any side effects
    """
    def __init__(self):
        Tool.__init__(self, "Crosshairs", "Navigate data without adding to ROI")
     
class PickedRegionTool(Tool):
    """
    Base class for tools which identify a picked region which can then be added
    erased or used as a mask
    """

    def __init__(self, name, desc, picker):
        Tool.__init__(self, name, desc)
        self._picker = picker

    def selected(self):
        Tool.selected(self)
        self.ivl.set_picker(self._picker)
   
    def interface(self):
        grid = Tool.interface(self)
        btn = QtWidgets.QPushButton("Add")
        btn.clicked.connect(self._change(self.builder.ADD))
        grid.addWidget(btn, 1, 0)
        btn = QtWidgets.QPushButton("Erase")
        btn.clicked.connect(self._change(self.builder.ERASE))
        grid.addWidget(btn, 1, 1)
        btn = QtWidgets.QPushButton("Mask")
        btn.clicked.connect(self._change(self.builder.MASK))
        grid.addWidget(btn, 2, 0)
        btn = QtWidgets.QPushButton("Discard")
        btn.clicked.connect(self.selected)
        grid.addWidget(btn, 2, 1)
        return grid

    def _change(self, mode):
        def _cb():
            selected_points = self.ivl.picker.selection(grid=self.builder.grid)
            if selected_points is not None:
                self.builder.modify(vol=selected_points, mode=mode)
                self.selected()
        return _cb

class PolygonTool(PickedRegionTool):
    """ Tool which selects a polygon region """
    def __init__(self):
        PickedRegionTool.__init__(self, "Polygon", "Select regions using a series of straight lines", PickMode.POLYGON)

class PenTool(PickedRegionTool):
    """ Tool which selects a region drawn around in freehand """
    def __init__(self):
        PickedRegionTool.__init__(self, "Pen", "Draw around ROI region", PickMode.FREEHAND)
     
class RectTool(PickedRegionTool):
    """ Tool which selects a rectangular region """
    def __init__(self):
        PickedRegionTool.__init__(self, "Rectangle", "Click and drag to select rectangular region", PickMode.RECT)

class EllipseTool(PickedRegionTool):
    """ Tool which selects an elliptical region """
    def __init__(self):
        PickedRegionTool.__init__(self, "Ellipse", "Click and drag to select elliptical region", PickMode.ELLIPSE)
       
class PainterTool(Tool):
    """
    Tool which paints voxels when clicked
    """
    def __init__(self, name="Painter", desc="Paint voxels on the ROI"):
        Tool.__init__(self, name, desc)
        self._paint_mode = 1 # FIXME

    def selected(self):
        self.ivl.set_picker(PickMode.PAINT)
        self.ivl.sig_selection_changed.connect(self._point_picked)

    def deselected(self):
        self.ivl.sig_selection_changed.disconnect(self._point_picked)
        Tool.deselected(self)

    def interface(self):
        grid = Tool.interface(self)
        self.size = NumericOption("Brush size", grid, 1, 0, intonly=True, maxval=10, default=1, minval=1, step=2)
        return grid

    def _point_picked(self, picker):
        self.debug("Point picked")
        size = self.size.spin.value()
        points = picker.selection(grid=self.builder.grid)

        grid_axes = self.builder.grid.get_ras_axes()
        gridx = grid_axes[picker.view.xaxis]
        gridy = grid_axes[picker.view.yaxis]
        gridz = grid_axes[picker.view.zaxis]
        self.debug("Grid axes: %i, %i, %i", gridx, gridy, gridz)
        self.debug(str(points))
        grid_points = []
        for point in points:
            for xd in range(-int((size-1)/2), int((size-1)/2)+1, 1):
                for yd in range(-int((size-1)/2), int((size-1)/2)+1, 1):
                    grid_point = [int(v + 0.5) for v in point]
                    grid_point[gridx] = grid_point[gridx] + xd
                    grid_point[gridy] = grid_point[gridy] + yd
                    if not grid_points or grid_point != grid_points[-1]:
                        grid_points.append(grid_point)
        self.debug(str(grid_points))
        self.builder.modify(points=grid_points, mode=self._paint_mode)
        picker.reset()
      
class EraserTool(PainterTool):
    """
    Tool which erases voxels when clicked
    """
    def __init__(self):
        PainterTool.__init__(self, "Eraser", "Remove voxels from the ROI")
        self._paint_mode = 2 # FIXME

class PickTool(Tool):
    """
    Tool which adds entire regions from another ROI
    """
    def __init__(self):
        Tool.__init__(self, "Pick", "Pick regions of an existing ROI")
        self.roi_name = ""
        self.temp_name = "PickToolTempRoi"

    def interface(self):
        grid = Tool.interface(self)

        grid.addWidget(QtWidgets.QLabel("Existing ROI"), 1, 0)
        self.roi_combo = RoiCombo(self.ivm, none_option=True)
        self.roi_combo.currentIndexChanged.connect(self._existing_roi_changed)
        grid.addWidget(self.roi_combo, 1, 1)

        self.ok_btn = QtWidgets.QPushButton("Accept")
        self.ok_btn.setEnabled(False)
        self.ok_btn.clicked.connect(self._accepted)
        grid.addWidget(self.ok_btn, 2, 0)
        self.cancel_btn = QtWidgets.QPushButton("Cancel")
        self.cancel_btn.setEnabled(False)
        self.cancel_btn.clicked.connect(self._reset)
        grid.addWidget(self.cancel_btn, 2, 1)
        self.done_btn = QtWidgets.QPushButton("Done")
        self.done_btn.setEnabled(False)
        self.done_btn.clicked.connect(self._show_builder_roi)
        grid.addWidget(self.done_btn, 2, 2)
        return grid

    def selected(self):
        self.ivl.set_picker(PickMode.SINGLE)
        self._show_existing_roi()
        self.ivl.sig_selection_changed.connect(self._point_picked)

    def deselected(self):
        self.ivl.sig_selection_changed.disconnect(self._point_picked)
        self._show_builder_roi()
        Tool.deselected(self)

    def _reset(self):
        self.roi_new = None
        self.ivm.delete(self.temp_name)
        self.ok_btn.setEnabled(False)
        self.cancel_btn.setEnabled(False)
        self.done_btn.setEnabled(True)
        self.roi_combo.setEnabled(True)
        self._show_builder_roi()

    def _show_existing_roi(self):
        if self.roi_name in self.ivm.rois:
            self.ivm.set_current_roi(self.roi_name)

    def _existing_roi_changed(self):
        self.roi_name = self.roi_combo.currentText()
        self._show_existing_roi()

    def _accepted(self):
        self.builder.modify(vol=self.roi_new, mode=self.builder.ADD)
        self._reset()

    def _point_picked(self, picker):
        if self.roi_name == "": return

        pos = picker.selection(grid=self.builder.grid)
        roi_picked = self.ivm.rois[self.roi_name]
        picked_region = roi_picked.value(pos, grid=self.builder.grid)

        roi_picked_arr = roi_picked.resample(self.builder.grid).raw()
        self.roi_new = np.zeros(self.builder.grid.shape, dtype=np.int)
        self.roi_new[roi_picked_arr == picked_region] = 1

        self.ivm.add(NumpyData(self.roi_new, grid=self.builder.grid, name=self.temp_name, roi=True), make_current=True)
        self.ok_btn.setEnabled(True)
        self.cancel_btn.setEnabled(True)
        self.done_btn.setEnabled(False)
        self.roi_combo.setEnabled(False)
        
class WalkerTool(Tool):
    """
    Tool which uses the random walker method to select a region

    FIXME this does not work properly in slice mode at the moment because the zaxis/pos 
    given by the picker does not necessarily correspond to the data axes.
    """
    def __init__(self):
        Tool.__init__(self, "Walker", "Automatic segmentation using the random walk algorithm")
        self.segmode = 0
        self.pickmode = 0

    def interface(self):
        grid = Tool.interface(self)

        grid.addWidget(QtWidgets.QLabel("Source data: "), 1, 0)
        self.ov_combo = OverlayCombo(self.ivm)
        grid.addWidget(self.ov_combo)

        grid.addWidget(QtWidgets.QLabel("Click to select points: "), 2, 0)
        self.pickmode_combo = QtWidgets.QComboBox()
        self.pickmode_combo.addItem("Inside the ROI")
        self.pickmode_combo.addItem("Outside the ROI")
        self.pickmode_combo.currentIndexChanged.connect(self._pick_mode_changed)
        grid.addWidget(self.pickmode_combo, 2, 1)
        
        grid.addWidget(QtWidgets.QLabel("Segmentation mode: "), 3, 0)
        self.segmode_combo = QtWidgets.QComboBox()
        self.segmode_combo.addItem("Slice")
        self.segmode_combo.addItem("3D")
        self.segmode_combo.currentIndexChanged.connect(self._seg_mode_changed)
        grid.addWidget(self.segmode_combo, 3, 1)

        self.beta = NumericOption("Diffusion difficulty", grid, 4, 0, intonly=True, maxval=20000, default=10000, step=1000)

        btn = QtWidgets.QPushButton("Segment")
        btn.clicked.connect(self._segment)
        grid.addWidget(btn, 5, 0)
        btn = QtWidgets.QPushButton("Clear points")
        btn.clicked.connect(self._init)
        grid.addWidget(btn, 5, 1)

        return grid

    def _pick_mode_changed(self, idx=None):
        self.pickmode = idx
        if self.pickmode == 0:
            self.ivl.picker.col = (255, 0, 0)
        else:
            self.ivl.picker.col = (255, 255, 255)

    def _seg_mode_changed(self, idx=None):
        self.segmode = idx
        self._init()

    def _init(self):
        if self.segmode == 1:
            self.ivl.set_picker(PickMode.MULTIPLE)
        else:
            self.ivl.set_picker(PickMode.SLICE_MULTIPLE)
            
        self.labels = np.zeros(self.builder.grid.shape, dtype=np.int)
        self._pick_mode_changed(self.pickmode)

    def selected(self):
        self.pickmode = 0
        self.segmode = 0
        self.ivl.sig_selection_changed.connect(self._points_changed)
        self._init()
        
    def deselected(self):
        self.ivl.sig_selection_changed.disconnect(self._points_changed)
        Tool.deselected(self)

    def _points_changed(self):
        for col, points in self.ivl.picker.selection(grid=self.builder.grid).items():
            if col == ((255, 0, 0)):
                label = 1
            else:
                label = 2

            for pos in points:
                pos = [int(p+0.5) for p in pos]
                # Clamp points to within range of ROI
                for dim in range(3):
                    pos[dim] = min(max(pos[dim], 0), self.labels.shape[dim]-1)
                self.labels[pos[0], pos[1], pos[2]] = label
                
    def _segment(self):
        data = self.ivm.data[self.ov_combo.currentText()].resample(self.builder.grid)
        labels = self.labels

        kwargs = {}
        # Use voxel size correctly
        spacing = [data.grid.spacing[0] / data.grid.spacing[0],
                   data.grid.spacing[0] / data.grid.spacing[1],
                   data.grid.spacing[0] / data.grid.spacing[2]]

        arr = data.raw()
        if arr.ndim > 3:
            # Reduce 4D data to PCA modes
            pca = PcaFeatReduce(n_components=5)
            arr = pca.get_training_features(arr, feature_volume=True)
            kwargs["multichannel"] = True
        else:
            # Normalize data
            arr = (arr / (np.max(arr)-np.min(arr))) + np.min(arr)
            kwargs["multichannel"] = False

        if self.segmode == 0:
            # Segment using 2D slice only
            zaxis = self.ivl.picker.zaxis
            zpos = int(self.ivl.picker.zpos + 0.5)
            #print(zaxis, zpos)
            sl = [slice(None)] * 3
            sl[zaxis] = zpos
            arr = arr[sl]
            labels = self.labels[sl]
            del spacing[zaxis] 
            #print(labels)
            #print(np.count_nonzero(labels))
            #print(np.count_nonzero(arr))

        seg = skimage.segmentation.random_walker(arr, labels, beta=self.beta.spin.value(), 
                                                 mode='cg_mg', spacing=spacing, **kwargs)

        if self.segmode == 0:
            # Create 3D volume using 2D slice
            seg_3d = np.zeros(self.builder.grid.shape, dtype=np.int)
            seg_3d[sl] = seg
            seg = seg_3d

        # Label 2 is used for 'outside region'
        seg[seg == 2] = 0
        
        self.builder.modify(vol=seg, mode=self.builder.ADD)
        self._init()
             
class BucketTool(Tool):
    """
    Tool which performs a thresholded bucket fill
    """

    def __init__(self):
        Tool.__init__(self, "Bucket", "2D or 3D flood fill with thresholding")
        self.point = None
        self.vol = 0

    def interface(self):
        grid = Tool.interface(self)

        #grid.addWidget(QtWidgets.QLabel("Segmentation mode: "), 3, 0)
        #self.segmode_combo = QtWidgets.QComboBox()
        #self.segmode_combo.addItem("Slice")
        #self.segmode_combo.addItem("3D")
        #self.segmode_combo.currentIndexChanged.connect(self._dims_changed)
        #grid.addWidget(self.segmode_combo, 3, 1)

        self.uthresh = NumericSlider("Upper threshold", grid, 4, 0, maxval=100, minval=0, default=50, hardmin=True)
        self.uthresh.sig_changed.connect(self._update_roi)
        self.lthresh = NumericSlider("Lower threshold", grid, 5, 0, maxval=0, minval=-100, default=-50, hardmax=True)
        self.lthresh.sig_changed.connect(self._update_roi)
        self.max_tile_size = NumericSlider("Max distance (voxels)", grid, 6, 0, intonly=True, maxval=500, minval=1, default=100, hardmin=True)
        self.max_tile_size.sig_changed.connect(self._update_roi)

        btn = QtWidgets.QPushButton("Add")
        btn.clicked.connect(self._add)
        grid.addWidget(btn, 7, 0)
        btn = QtWidgets.QPushButton("Erase")
        btn.clicked.connect(self._erase)
        grid.addWidget(btn, 7, 1)
        btn = QtWidgets.QPushButton("Mask")
        btn.clicked.connect(self._mask)
        grid.addWidget(btn, 8, 0)
        btn = QtWidgets.QPushButton("Discard")
        btn.clicked.connect(self._init)
        grid.addWidget(btn, 8, 1)
        return grid

    def _init(self):
        self.ivl.set_picker(PickMode.SINGLE)
        self._point = None
        if "_temp_bucket" in self.ivm.rois:
            self.ivm.delete("_temp_bucket")
        self._show_builder_roi()

    def selected(self):
        self.ivl.sig_selection_changed.connect(self._sel_changed)
        self._init()
        self._sel_changed()
        
    def deselected(self):
        self.ivl.sig_selection_changed.disconnect(self._sel_changed)
        Tool.deselected(self)
        self._init()

    def _sel_changed(self):
        focus = self.ivl.picker.selection(grid=self.builder.grid)
        self.point = [int(v+0.5) for v in focus[:3]]
        self.vol = focus[3]
        self._update_roi()
    
    def _update_roi(self):
        src_data = self.ivm.current_data.resample(self.builder.grid).volume(self.vol)
        focus_value = src_data[self.point[0], self.point[1], self.point[2]]
        thr_hi, thr_lo = focus_value + self.uthresh.value(), focus_value + self.lthresh.value()
        # Heuristic optimization for large data sets. Start by looking at a tile +- 50 voxels
        # around focus. If that contains the region, fine. If not, add 50 voxels to the tile
        # size and go again until we get nothing on the boundary or end up taking the whole
        # data set
        max_tile_size = self.max_tile_size.value()
        tile_size = min(50, max_tile_size)
        while 1:
            tile, offset = self._get_tile(src_data, self.point, tile_size, src_data.shape)
            binarised = ((tile <= thr_hi) & (tile >= thr_lo)).astype(np.int)
            labelled, _ = scipy.ndimage.measurements.label(binarised)
            scipy_label = labelled[self.point[0]-offset[0], self.point[1]-offset[1], self.point[2]-offset[2]]
            labelled[labelled != scipy_label] = 0
            labelled[labelled == scipy_label] = 1
            if labelled.shape == src_data.shape or tile_size == max_tile_size:
                self.debug("Reached full size, breaking: ", labelled.shape, src_data.shape)
                break
            elif (np.count_nonzero(labelled[:, :, 0]) == 0 and 
                  np.count_nonzero(labelled[:, :, -1]) == 0 and
                  np.count_nonzero(labelled[:, 0, :]) == 0 and
                  np.count_nonzero(labelled[:, -1, :]) == 0 and
                  np.count_nonzero(labelled[0, :, :]) == 0 and
                  np.count_nonzero(labelled[-1, :, :]) == 0):
                self.debug("Nothing on boundary, breaking")
                break
            tile_size = min(tile_size + 50, max_tile_size)

        self.roi = np.zeros(self.builder.grid.shape, dtype=np.int)
        tile_shape = labelled.shape
        self.roi[offset[0]:offset[0]+tile_shape[0], offset[1]:offset[1]+tile_shape[1], offset[2]:offset[2]+tile_shape[2]] = labelled
        self.ivm.add(self.roi, name="_temp_bucket", grid=self.builder.grid, roi=True, make_current=True)
        
    def _get_tile(self, arr, centre, size, shape):
        slices, offset = [], []
        for i in range(3):
            start, end = centre[i]-size, centre[i]+size
            if start < 0: start = 0
            if end > shape[i]: end = shape[i]
            slices.append(slice(start, end))
            offset.append(start)
        self.debug("Tile: ", slices, offset)
        return arr[slices], offset

    def _add(self):
        self.builder.modify(vol=self.roi, mode=self.builder.ADD)
        self._init()

    def _erase(self):
        self.builder.modify(vol=self.roi, mode=self.builder.ERASE)
        self._init()

    def _mask(self):
        self.builder.modify(vol=self.roi, mode=self.builder.MASK)
        self._init()

    def _dims_changed(self):
        pass



"""
Quantiphyse - Widgets for comparing data sets

Copyright (c) 2013-2020 University of Oxford

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from PySide2 import QtGui, QtCore, QtWidgets
  
import pyqtgraph as pg
import numpy as np

from quantiphyse.gui.widgets import QpWidget, TitleWidget, OverlayCombo, NumericOption, ChoiceOption
from quantiphyse.utils import QpException

class CompareDataWidget(QpWidget):
    """
    Compare two data sets
    """
    def __init__(self, **kwargs):
        QpWidget.__init__(self, name="Compare Data", icon="compare.png", 
                          desc="Compare two data sets", group="Visualisation", **kwargs)
        
    def init_ui(self):
        vbox = QtWidgets.QVBoxLayout()
        self.setLayout(vbox)

        title = TitleWidget(self, batch_btn=False, help="compare")
        vbox.addWidget(title)

        hbox = QtWidgets.QHBoxLayout()
        hbox.addWidget(QtWidgets.QLabel("Compare "))
        self.d1_combo = OverlayCombo(self.ivm)
        self.d1_combo.currentIndexChanged.connect(self._update_data)
        hbox.addWidget(self.d1_combo)
        hbox.addWidget(QtWidgets.QLabel(" with "))
        self.d2_combo = OverlayCombo(self.ivm)
        self.d2_combo.currentIndexChanged.connect(self._update_data)
        hbox.addWidget(self.d2_combo)
        hbox.addStretch(1)
        vbox.addLayout(hbox)

        hbox = QtWidgets.QHBoxLayout()
        hbox.addWidget(QtWidgets.QLabel("Within ROI "))
        self.roi_combo = OverlayCombo(self.ivm, rois=True, data=False, none_option=True)
        self.roi_combo.currentIndexChanged.connect(self._update_data)
        hbox.addWidget(self.roi_combo)
        hbox.addStretch(1)
        vbox.addLayout(hbox)

        hbox = QtWidgets.QHBoxLayout()
        gbox = QtWidgets.QGroupBox("Options")
        grid = QtWidgets.QGridLayout()
        gbox.setLayout(grid)

        self.id_cb = QtWidgets.QCheckBox("Include identity line")
        grid.addWidget(self.id_cb, 0, 0)
        self.sample_cb = QtWidgets.QCheckBox("Sample values")
        self.sample_cb.setChecked(True)
        self.sample_cb.stateChanged.connect(self._update_data)
        self.sample_cb.stateChanged.connect(self._update_gui)
        grid.addWidget(self.sample_cb, 1, 0)
        self.sample_spin = QtWidgets.QSpinBox()
        self.sample_spin.setMinimum(10)
        self.sample_spin.setMaximum(10000000)
        self.sample_spin.setSingleStep(100)
        self.sample_spin.setValue(1000)
        self.sample_spin.valueChanged.connect(self._update_data)
        grid.addWidget(self.sample_spin, 1, 1)
        self.warn_label = QtWidgets.QLabel("WARNING: plotting all values may take a long time")
        self.warn_label.setStyleSheet("QLabel { color : red; }")
        self.warn_label.setVisible(False)
        grid.addWidget(self.warn_label, 2, 0)
        self.plot_mode = ChoiceOption("Plot mode", grid, 3, choices=["Scatter", "Heat map"])
        self.plot_mode.combo.currentIndexChanged.connect(self._plot_mode_changed)
        self.bins = NumericOption("Bins", grid, 3, xpos=2, minval=20, default=50, intonly=True)
        self.bins.label.setVisible(False)
        self.bins.spin.setVisible(False)
        self.run_btn = QtWidgets.QPushButton("Update")
        self.run_btn.clicked.connect(self._run)
        grid.addWidget(self.run_btn, 4, 0)
        hbox.addWidget(gbox)
        hbox.addStretch(1)
        vbox.addLayout(hbox)

        hbox = QtWidgets.QHBoxLayout()
        win = pg.GraphicsLayoutWidget()
        win.setBackground(background=None)
        self.plot = win.addPlot(enableAutoRange=True)
        self.plot.clear() # Unsure why this is necessary
        hbox.addWidget(win)
        
        self.hist = pg.HistogramLUTWidget()
        self.hist.setVisible(False)
        self.hist.setBackground(None)
        self.hist.gradient.loadPreset("thermal")
        hbox.addWidget(self.hist) 
        vbox.addLayout(hbox)

        self._update_gui()
        self._update_data()
    
    def _plot_mode_changed(self, idx):
        self.bins.label.setVisible(idx == 1)
        self.bins.spin.setVisible(idx == 1)
        self.sample_cb.setChecked(idx == 0)
        self._update_gui()
        self._update_data()

    def _update_gui(self):
        self.sample_spin.setEnabled(self.sample_cb.isChecked())

    def _update_data(self):
        name1 = self.d1_combo.currentText()
        name2 = self.d2_combo.currentText()
        roi_name = self.roi_combo.currentText()
        qpd1 = self.ivm.data.get(name1, None)
        qpd2 = self.ivm.data.get(name2, None)
        roi = self.ivm.rois.get(roi_name, None)

        if qpd1 is not None and qpd2 is not None:
            current_vol = self.ivl.focus()[3]
            d1 = qpd1.volume(current_vol)
            d2 = qpd2.resample(qpd1.grid).volume(current_vol)
            
            if roi is not None:
                roi_data = roi.resample(qpd1.grid).raw()
                d1 = d1[roi_data > 0]
                d2 = d2[roi_data > 0]
            else:
                d1 = d1.reshape(-1)
                d2 = d2.reshape(-1)

            if self.sample_cb.isChecked():
                n_samples = self.sample_spin.value()
                idx = np.random.choice(np.arange(len(d1)), n_samples)
                d1 = np.take(d1, idx)
                d2 = np.take(d2, idx)
                self.warn_label.setVisible(False)
            else:
                # Warn about plotting all data in scatter mode
                self.warn_label.setVisible(self.plot_mode.combo.currentIndex() == 0)
                        
            self.d1 = d1
            self.d2 = d2
            self.run_btn.setEnabled(True)
        else:
            self.run_btn.setEnabled(False)

    def _run(self):
        self.plot.clear() 
        self.plot.setLabel('bottom', self.d1_combo.currentText())
        self.plot.setLabel('left', self.d2_combo.currentText())
        if self.plot_mode.combo.currentIndex() == 0:
            self.plot.setAspectLocked(False)
            self.hist.setVisible(False)
            self.plot.plot(self.d1, self.d2, pen=None, symbolBrush=(200, 200, 200), symbolPen='k', symbolSize=5.0)
        else:
            heatmap, xedges, yedges = np.histogram2d(self.d1, self.d2, bins=self.bins.value())
            rect = QtCore.QRectF(xedges[0], yedges[0], xedges[-1]-xedges[0], yedges[-1]-yedges[0])
            img = pg.ImageItem(heatmap)
            img.setRect(rect)
            #self.plot.setAspectLocked(True)
            self.hist.setVisible(True)
            self.hist.setImageItem(img)
            self.plot.addItem(img)

        if self.id_cb.isChecked():
            real_min = max(self.d1.min(), self.d2.min())
            real_max = min(self.d1.max(), self.d2.max())
            pen = pg.mkPen((255, 255, 255), style=QtCore.Qt.DashLine)
            self.plot.plot([real_min, real_max], [real_min, real_max], pen=pen, width=2.0)



"""
Quantiphyse - PCA reduction process

Copyright (c) 2013-2020 University of Oxford

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import numpy as np

from quantiphyse.data.extras import MatrixExtra
from quantiphyse.utils import QpException
from quantiphyse.processes import Process
from quantiphyse.processes.feat_pca import PcaFeatReduce

class PcaProcess(Process):
    """
    Process to do PCA (Principal Component Analysis) reduction on 4D data
    """
    PROCESS_NAME = "PCA"

    def __init__(self, ivm, **kwargs):
        Process.__init__(self, ivm, **kwargs)
        self.explained_variance = []
        self.pca_modes = []
        self.mean = [0,]

    def run(self, options):
        data = self.get_data(options)
        roi = self.get_roi(options, data.grid)
        output_name = options.pop("output-name", "%s_pca" % data.name)
        norm_input = options.pop('norm-input', True)
        norm_type = options.pop('norm-type', "sigenh")
        norm_output = options.pop('norm-output', False)
        n_components = options.pop('n-components', 5)

        if data.ndim != 4:
            raise QpException("PCA reduction possible on 4D data only")
        elif data.nvols <= n_components:
            raise QpException("Number of PCA components must be less than number of data volumes")

        pca = PcaFeatReduce(n_components=n_components, norm_input=norm_input, norm_type=norm_type, norm_modes=norm_output)
        
        feature_images = pca.get_training_features(data.raw(), roi.raw(), feature_volume=True)
        for comp_idx in range(n_components):
            name = "%s%i" % (output_name, comp_idx)
            self.ivm.add(feature_images[:, :, :, comp_idx], grid=data.grid, name=name, make_current=(comp_idx == 0))

        #if options.pop("reduced-data", True):
        #    reduced = pca.pca.inverse_transform(feature_images)
        #    self.ivm.add(reduced, grid=data.grid, name=name + "_reduced")

        cumulative = 0
        var_rows = []
        for idx, variance in enumerate(pca.explained_variance()):
            cumulative += variance
            var_rows.append([idx, variance, cumulative])
        extra = MatrixExtra(output_name + "_variance", var_rows, col_headers=["PCA mode", "Explained variance", "Cumulative"])
        self.ivm.add_extra(extra.name, extra)

        modes = np.zeros((len(pca.mean()), len(pca.modes()) + 1))
        cols = []
        for idx, mode in enumerate(pca.modes()):
            modes[:, idx] = mode
            cols.append("Mode %i" % idx)
        modes[:, -1] = pca.mean()
        cols.append("Mean")
        extra = MatrixExtra(output_name + "_modes", modes, col_headers=cols)
        self.ivm.add_extra(extra.name, extra)



print("import: 'PySide2'")
import PySide2

print("import: 'quantiphyse'")
import quantiphyse

print("import: 'quantiphyse.data'")
import quantiphyse.data

print("import: 'quantiphyse.gui'")
import quantiphyse.gui

print("import: 'quantiphyse.icons'")
import quantiphyse.icons

print("import: 'quantiphyse.packages'")
import quantiphyse.packages

print("import: 'quantiphyse.packages.core'")
import quantiphyse.packages.core

print("import: 'quantiphyse.packages.core.analysis'")
import quantiphyse.packages.core.analysis

print("import: 'quantiphyse.packages.core.batch_builder'")
import quantiphyse.packages.core.batch_builder

print("import: 'quantiphyse.packages.core.clustering'")
import quantiphyse.packages.core.clustering

print("import: 'quantiphyse.packages.core.compare'")
import quantiphyse.packages.core.compare

print("import: 'quantiphyse.packages.core.data'")
import quantiphyse.packages.core.data

print("import: 'quantiphyse.packages.core.experimental'")
import quantiphyse.packages.core.experimental

print("import: 'quantiphyse.packages.core.histogram'")
import quantiphyse.packages.core.histogram

print("import: 'quantiphyse.packages.core.overview'")
import quantiphyse.packages.core.overview

print("import: 'quantiphyse.packages.core.pca'")
import quantiphyse.packages.core.pca

print("import: 'quantiphyse.packages.core.radialprofile'")
import quantiphyse.packages.core.radialprofile

print("import: 'quantiphyse.packages.core.registration'")
import quantiphyse.packages.core.registration

print("import: 'quantiphyse.packages.core.roi_builder'")
import quantiphyse.packages.core.roi_builder

print("import: 'quantiphyse.packages.core.simulation'")
import quantiphyse.packages.core.simulation

print("import: 'quantiphyse.packages.core.smoothing'")
import quantiphyse.packages.core.smoothing

print("import: 'quantiphyse.processes'")
import quantiphyse.processes

print("import: 'quantiphyse.resources'")
import quantiphyse.resources

print("import: 'quantiphyse.test'")
import quantiphyse.test

print("import: 'quantiphyse.utils'")
import quantiphyse.utils




"""
Quantiphyse - Widgets to manipulate point of focus

Copyright (c) 2013-2020 University of Oxford

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from __future__ import division, unicode_literals, absolute_import

from PySide2 import QtGui, QtCore, QtWidgets

from quantiphyse.utils import LogSource
from quantiphyse.gui.widgets import OptionsButton

class Navigator(LogSource):
    """
    Slider control which alters position along an axis
    """

    def __init__(self, ivl, label, axis, layout_grid, layout_ypos):
        LogSource.__init__(self)
        self.ivl = ivl
        self.axis = axis
        self.data_axis = axis
        self.data_grid = None
        self._pos = -1

        layout_grid.addWidget(QtWidgets.QLabel(label), layout_ypos, 0)
        self.slider = QtWidgets.QSlider(QtCore.Qt.Horizontal)
        self.slider.setFocusPolicy(QtCore.Qt.NoFocus)
        self.slider.setMinimumWidth(100)
        self.slider.valueChanged.connect(self._changed)
        layout_grid.addWidget(self.slider, layout_ypos, 1)

        self.spin = QtWidgets.QSpinBox()
        self.spin.valueChanged.connect(self._changed)
        layout_grid.addWidget(self.spin, layout_ypos, 2)

        self.ivl.ivm.sig_main_data.connect(self._main_data_changed)
        self.ivl.sig_focus_changed.connect(self._focus_changed)

    def _changed(self, value):
        if value != self._pos and self.data_grid is not None:
            pos = self.ivl.focus(self.data_grid)
            pos[self.data_axis] = value
            self.ivl.set_focus(pos, self.data_grid)

    def _main_data_changed(self, data):
        if data is not None:
            self.data_grid = data.grid
            self.data_axis = data.grid.get_ras_axes()[self.axis]

            if self.axis < 3:
                self._set_size(self.data_grid.shape[self.data_axis])
            else:
                self._set_size(data.nvols)

            self._focus_changed()
        else:
            self.data_grid = None
            self.data_axis = self.axis
            self._set_size(1)
            self._pos = 0

    def _focus_changed(self):
        if self.data_grid is not None:
            self._pos = int(self.ivl.focus(self.data_grid)[self.data_axis]+0.5)
            self.debug("Pos for slider %i %i", self.axis, self._pos)
            try:
                self.slider.blockSignals(True)
                self.spin.blockSignals(True)
                self.slider.setValue(self._pos)
                self.spin.setValue(self._pos)
            finally:
                self.slider.blockSignals(False)
                self.spin.blockSignals(False)

    def _set_size(self, size):
        try:
            self.slider.blockSignals(True)
            self.spin.blockSignals(True)
            self.slider.setRange(0, size-1)
            self.spin.setMaximum(size-1)
        finally:
            self.slider.blockSignals(False)
            self.spin.blockSignals(False)

class VolumeNavigator(Navigator):
    """
    Slider navigator control specifically for the volume axis
    """

    def __init__(self, *args, **kwargs):
        Navigator.__init__(self, label="Volume", axis=3, *args, **kwargs)
        self.ivl.ivm.sig_all_data.connect(self._data_changed)

    def _main_data_changed(self, data):
        if data is not None:
            self.data_grid = data.grid
        self._data_changed()

    def _data_changed(self):
        max_num_vols = max([d.nvols for d in self.ivl.ivm.data.values()] + [1, ])
        self._set_size(max_num_vols)
        self._focus_changed()

class NavigationBox(QtWidgets.QGroupBox):
    """
    Box containing 4D navigators
    """
    def __init__(self, ivl):
        QtWidgets.QGroupBox.__init__(self)
        self.ivl = ivl

        grid = QtWidgets.QGridLayout()
        grid.setVerticalSpacing(2)
        grid.setContentsMargins(5, 5, 5, 5)
        self.setLayout(grid)

        self.navs = []
        self.navs.append(Navigator(ivl, "Axial", 2, grid, 0))
        self.navs.append(Navigator(ivl, "Sagittal", 0, grid, 1))
        self.navs.append(Navigator(ivl, "Coronal", 1, grid, 2))
        self.navs.append(VolumeNavigator(ivl, layout_grid=grid, layout_ypos=3))
        grid.setColumnStretch(0, 0)
        grid.setColumnStretch(1, 2)



"""
Quantiphyse - Processes for the data manipulation package

Copyright (c) 2013-2020 University of Oxford

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import math

import numpy as np
import scipy.ndimage

from quantiphyse.data import NumpyData, DataGrid
from quantiphyse.utils import QpException
from quantiphyse.processes import Process

class ResampleProcess(Process):
    """ 
    Resample data 
    """

    PROCESS_NAME = "Resample"
    
    def run(self, options):
        data = self.get_data(options)
        if data.roi: 
            default_order=0
        else:
            default_order=1
        order = options.pop("order", default_order)
        resample_type = options.pop("type", "data")
        output_name = options.pop("output-name", "%s_res" % data.name)
        grid_data = options.pop("grid", None)
        factor = options.pop("factor", None)
        only2d = options.pop("2d", None)

        # The different types of resampling require significantly different strategies
        #
        # Data->Data resampling is implemented in the QpData class although this will give
        # results which are not ideal when resampling to a lower resolution.
        # Upsampling can use scipy.ndimage.zoom
        # Downsampling is nore naturally implemented as a mean over subvoxels using Numpy slicing
        #
        # Note that factor is an integer for now. It could easily be a float for upsampling but
        # this would break the downsampling algorithm (and make it significanlty more complex to
        # implement)
        #
        # This is all pretty messy now especially with the '2d only' option.
        if resample_type == "data":
            if grid_data is None:
                raise QpException("Must provide 'grid' option to specify data item to get target grid from")
            elif grid_data not in self.ivm.data:
                raise QpException("Data item '%s' not found" % grid_data)
            
            grid = self.ivm.data[grid_data].grid
            output_data = data.resample(grid, order=order)
        elif resample_type == "up":
            # Upsampling will need to use interpolation
            orig_data = data.raw()
            zooms = [factor for idx in range(3)]
            if only2d:
                zooms[2] = 1
            if data.ndim == 4:
                zooms.append(1)
            output_data = scipy.ndimage.zoom(orig_data, zooms, order=order)

            # Work out new grid origin
            voxel_offset = [float(factor-1)/(2*factor) for idx in range(3)]
            if only2d:
                voxel_offset[2] = 0
            offset = data.grid.grid_to_world(voxel_offset, direction=True)
            output_affine = np.array(data.grid.affine)
            for idx in range(3):
                if idx < 2 or not only2d:
                    output_affine[:3, idx] /= factor
            output_affine[:3, 3] -= offset

            output_grid = DataGrid(output_data.shape[:3], output_affine)
            output_data = NumpyData(output_data, grid=output_grid, name=output_name)
        elif resample_type == "down":
            # Downsampling takes a mean of the voxels inside the new larger voxel
            # Only uses integral factor at present
            orig_data = data.raw()
            new_shape = [max(1, int(dim_size / factor)) for  dim_size in orig_data.shape[:3]]
            if data.ndim == 4:
                new_shape.append(orig_data.shape[3])
            if only2d:
                new_shape[2] = orig_data.shape[2]

            # Note that output data must be float data type even if original data was integer
            output_data = np.zeros(new_shape, dtype=np.float32)
            num_samples = 0
            for start1 in range(factor):
                for start2 in range(factor):
                    for start3 in range(factor):
                        if start1 >= new_shape[0]*factor or start2 >= new_shape[1]*factor or start3 >= new_shape[2]*factor:
                            continue
                        slices = [
                            slice(start1, new_shape[0]*factor, factor),
                            slice(start2, new_shape[1]*factor, factor),
                            slice(start3, new_shape[2]*factor, factor),
                        ]
                        if only2d:
                            slices[2] = slice(None)
                        downsampled_data=orig_data[slices]
                        output_data += downsampled_data
                        num_samples += 1
            output_data /= num_samples
            # FIXME this will not work for 2D data
            voxel_offset = [0.5*(factor-1), 0.5*(factor-1), 0.5*(factor-1)]
            if only2d:
                voxel_offset[2] = 0
            offset = data.grid.grid_to_world(voxel_offset, direction=True)
            output_affine = np.array(data.grid.affine)
            for idx in range(3):
                if idx < 2 or not only2d:
                    output_affine[:3, idx] *= factor
            output_affine[:3, 3] += offset

            output_grid = DataGrid(output_data.shape[:3], output_affine)
            output_data = NumpyData(output_data, grid=output_grid, name=output_name)
        elif resample_type == "res":
            # Resampling to specified resolution
            voxel_sizes = options.pop("voxel-sizes", None)
            self.debug("Voxel sizes: %s", voxel_sizes)
            if voxel_sizes is None:
                raise QpException("Must specifiy voxel sizes")
            src_grid = data.grid
            self.debug("Current voxel sizes: %s", src_grid.spacing)
            scale_factors = [src_grid.spacing[dim] / voxel_sizes[dim] for dim in range(3)]
            self.debug("Scale factors: %s", scale_factors)
            new_shape = [int(math.ceil(src_grid.shape[dim] * scale_factors[dim])) for dim in range(3)]
            self.debug("New shape is %s", new_shape)
            self.debug("Input affine: %s", src_grid.affine)
            output_affine = np.copy(src_grid.affine)
            for dim in range(3):
                output_affine[:, dim] /= scale_factors[dim]
            self.debug("Output affine: %s", output_affine)
            output_grid = DataGrid(new_shape, output_affine)
            output_data = data.resample(output_grid, order=order)
        else:
            raise QpException("Unknown resampling type: %s" % resample_type)

        self.ivm.add(output_data, name=output_name, make_current=True, roi=data.roi and order == 0)



"""
Quantiphyse - Base class for logging

Copyright (c) 2013-2020 University of Oxford

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import logging
logging.basicConfig(format='%(asctime)s %(levelname)s %(name)s %(message)s', datefmt='%H:%M:%S')

def set_base_log_level(level):
    """
    Set the base logging level
    """
    logging.getLogger("quantiphyse").setLevel(level)

class LogSource(object):
    """
    Base class for anything which wants to log messages
    """
    def __init__(self):
        logname = "%s.%s" % (self.__module__, self.__class__.__name__)
        if not logname.startswith("quantiphyse."):
            # Plugins do not come under the quantiphyse namespace but we want them
            # to be ancestors of the generic logger
            logname = "quantiphyse." + logname
        self.logger = logging.getLogger(logname)

    def debug(self, *args, **kwargs):
        """
        Log a debug level message
        """
        self.logger.debug(*args, **kwargs)

    def warn(self, *args, **kwargs):
        """
        Log a warning
        """
        self.logger.warn(*args, **kwargs)
    
    def debug_enabled(self):
        return self.logger.getEffectiveLevel() <= logging.DEBUG


"""
Quantiphyse - Dialog box for editing view options

Copyright (c) 2013-2020 University of Oxford

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from __future__ import division, unicode_literals, print_function, absolute_import

from PySide2 import QtGui, QtCore, QtWidgets

from quantiphyse.utils.enums import Orientation, DisplayOrder, Visibility
from quantiphyse.gui.options import OptionBox, ChoiceOption

class ViewerOptions(QtWidgets.QDialog):
    """
    Dialog box which controls viewer options
    """

    def __init__(self, parent, ivl):
        super(ViewerOptions, self).__init__(parent)
        self.setWindowTitle("View Options")
        self._ivl = ivl

        vbox = QtWidgets.QVBoxLayout()
        label = QtWidgets.QLabel('<font size="5">View Options</font>')
        vbox.addWidget(label)

        self._optbox = OptionBox()
        self._optbox.add("Orientation", 
                         ChoiceOption(["Radiological (Right is Left)", "Neurological (Left is Left)"],
                                      [Orientation.RADIOLOGICAL, Orientation.NEUROLOGICAL], 
                                      default=self._ivl.opts.orientation),
                         key="orient")
        self._optbox.add("Crosshairs", 
                         ChoiceOption(["Show", "Hide"], [Visibility.SHOW, Visibility.HIDE],
                                      default=self._ivl.opts.crosshairs),
                         key="crosshairs")
        self._optbox.add("Orientation labels", 
                         ChoiceOption(["Show", "Hide"], [Visibility.SHOW, Visibility.HIDE],
                                      default=self._ivl.opts.labels),
                         key="labels")
        self._optbox.add("Main data background", 
                         ChoiceOption(["Show", "Hide"], [Visibility.SHOW, Visibility.HIDE],
                                      default=self._ivl.opts.main_data),
                         key="main_data")
        self._optbox.add("Display order", 
                         ChoiceOption(["User selected", "Data always on top", "ROIs always on top"], 
                                      [DisplayOrder.USER, DisplayOrder.DATA_ON_TOP, DisplayOrder.ROI_ON_TOP],
                                      default=self._ivl.opts.display_order),
                         key="display_order")
        self._optbox.add("View interpolation", 
                         ChoiceOption(["Nearest neighbour (fast)", "Linear", "Cubic spline (slow)"], [0, 1, 3],
                                      default=self._ivl.opts.interp),
                         key="interp")
        self._optbox.option("orient").sig_changed.connect(self._orientation_changed)
        self._optbox.option("crosshairs").sig_changed.connect(self._crosshairs_changed)
        self._optbox.option("labels").sig_changed.connect(self._labels_changed)
        self._optbox.option("main_data").sig_changed.connect(self._main_data_changed)
        self._optbox.option("display_order").sig_changed.connect(self._display_order_changed)
        self._optbox.option("interp").sig_changed.connect(self._interp_changed)
        vbox.addWidget(self._optbox)

        vbox.addStretch(1)
        self.setLayout(vbox)

    def _orientation_changed(self):
        self._ivl.opts.orientation = self._optbox.option("orient").value

    def _crosshairs_changed(self):
        self._ivl.opts.crosshairs = self._optbox.option("crosshairs").value

    def _labels_changed(self):
        self._ivl.opts.labels = self._optbox.option("labels").value

    def _main_data_changed(self):
        self._ivl.opts.main_data = self._optbox.option("main_data").value

    def _display_order_changed(self):
        self._ivl.opts._display_order = self._optbox.option("display_order").value

    def _interp_changed(self):
        self._ivl.opts.interp = self._optbox.option("interp").value



"""
Quantiphyse - AIF widget

Copyright (c) 2013-2020 University of Oxford

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
from PySide2 import QtGui, QtCore, QtWidgets

import numpy as np

from quantiphyse.data.extras import NumberListExtra
from quantiphyse.gui.widgets import QpWidget, TitleWidget
from quantiphyse.gui.dialogs import TextViewerDialog
from quantiphyse.gui.pickers import PickMode
from quantiphyse.gui.plot import Plot
from quantiphyse.gui.options import OptionBox, DataOption, NumericOption, BoolOption, ChoiceOption, TextOption

class AifWidget(QpWidget):
    """
    Widget which allows the user to define an arterial input function from signal data
    """
    def __init__(self, **kwargs):
        super(AifWidget, self).__init__(name="Arterial Input Function", icon="aif",
                                        desc="Tool for defining an AIF from measured data", group="Utilities", **kwargs)
        self._aif = []
        self._aif_points = []

    def init_ui(self):
        vbox = QtWidgets.QVBoxLayout()
        self.setLayout(vbox)

        title = TitleWidget(self)
        vbox.addWidget(title)

        self._options = OptionBox("Options")
        self._options.add("Signal data", DataOption(self.ivm), key="data")
        self._clear_btn = QtWidgets.QPushButton("Clear points")
        self._options.add("Method", ChoiceOption(["Pick points", "Use existing ROI"], ["points", "roi"]), self._clear_btn, key="method")
        self._options.add("ROI", DataOption(self.ivm, data=False, rois=True), key="roi")
        self._view_btn = QtWidgets.QPushButton("View")
        self._view_btn.setEnabled(False)
        self._save_btn = QtWidgets.QPushButton("Save")
        self._save_btn.setEnabled(False)
        self._options.add("AIF name", TextOption("aif"), self._view_btn, self._save_btn, key="output-name")
        self._options.option("method").sig_changed.connect(self._method_changed)
        self._options.option("data").sig_changed.connect(self._recalc_aif)
        self._options.option("roi").sig_changed.connect(self._recalc_aif)
        self._clear_btn.clicked.connect(self._clear_btn_clicked)
        self._save_btn.clicked.connect(self._save_btn_clicked)
        self._view_btn.clicked.connect(self._view_btn_clicked)
        vbox.addWidget(self._options)

        self._plot = Plot(qpo=None, parent=self, title="AIF", display_mode=False)
        self._plot.set_xlabel("Volume")
        self._plot.set_ylabel("Signal")
        vbox.addWidget(self._plot)

        vbox.addStretch(1)

    def activate(self):
        self.ivl.sig_selection_changed.connect(self._selection_changed)
        self._method_changed()

    def deactivate(self):
        self.ivl.sig_selection_changed.disconnect(self._selection_changed)
        self.ivl.set_picker(PickMode.SINGLE)

    @property
    def qpdata(self):
        return self.ivm.data.get(self._options.option("data").value, None)

    @property
    def roi(self):
        return self.ivm.data.get(self._options.option("roi").value, None)

    @property
    def method(self):
        return self._options.option("method").value

    def _method_changed(self):
        self._options.set_visible("roi", self.method == "roi")
        self._clear_btn.setVisible(self.method == "points")
        if self.method == "roi":
            self.ivl.set_picker(PickMode.SINGLE)
        else:
            self.ivl.set_picker(PickMode.MULTIPLE)
            self._selection_changed()
        self._recalc_aif()

    def _clear_btn_clicked(self):
        if self.method == "points":
            self.ivl.set_picker(PickMode.MULTIPLE)
            self._selection_changed() # FIXME should be signalled by picker

    def _save_btn_clicked(self):
        name = self._options.option("output-name").value
        extra = NumberListExtra(name, self._aif)
        self.ivm.add_extra(name, extra)
        self._save_btn.setEnabled(False)

    def _view_btn_clicked(self):
        aiftxt = ", ".join([str(v) for v in self._aif])
        TextViewerDialog(self, title="AIF data", text=aiftxt).exec_()

    def _selection_changed(self):
        if self.method == "roi":
            return
        
        self._aif_points = []
        for _col, points in self.ivl.picker.selection().items():
            self._aif_points += list(points)
        self._recalc_aif()

    def _recalc_aif(self):
        self._aif = []
        self._save_btn.setEnabled(True)
        self._view_btn.setEnabled(True)
        if self.qpdata is not None:
            if self.method == "roi":
                self._calc_aif_roi()
            else:
                self._calc_aif_points()
        self._update_plot()

    def _calc_aif_roi(self):
        if self.roi is None:
            return

        points = self.qpdata.raw()[self.roi.raw() > 0]
        if len(points) > 0:
            aif = None
            for sig in points:
                if aif is None:
                    aif = np.zeros([len(sig)], dtype=np.float32)
                aif += sig
            self._aif = aif / len(points)

    def _calc_aif_points(self):
        aif = None
        num_points = 0
        for point in self._aif_points:
            sig = self.qpdata.timeseries(point, grid=self.ivl.grid)
            self.debug("AIF signal: %s", sig)
            if aif is None:
                aif = np.zeros([len(sig)], dtype=np.float32)
            aif += sig
            num_points += 1

        if num_points > 0:
            self._aif = aif / num_points

    def _update_plot(self):
        self._plot.clear()
        self._plot.add_line(self._aif, name="AIF")



"""
Quantiphyse - Custom dialog boxes

Copyright (c) 2013-2020 University of Oxford

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import os

from PySide2 import QtGui, QtCore, QtWidgets

MAINWIN = None

def set_main_window(win):
    """ Set the main window widget so it can be a parent to dialogs to make style match"""
    global MAINWIN
    MAINWIN = win

def error_dialog(msg, title="Warning", detail=None, subtitle="Details:"):
    """
    Show an error dialog box
    """
    text = msg.replace(os.linesep, "<br>")
    if detail is not None:
        detail_str = ""
        if isinstance(detail, (list, tuple)):
            for item in detail:
                detail_str += str(item) + os.linesep
        else:
            detail_str = str(detail)
        detail_str = detail_str.replace(os.linesep, "<br>")
        text += "<br><br><b>%s</b><br><br>%s" % (subtitle, detail_str)

    QtWidgets.QMessageBox.warning(MAINWIN, title, text, QtWidgets.QMessageBox.Close)

class MultiTextViewerDialog(QtWidgets.QDialog):
    """
    Text viewer dialog with multiple pages presented as tabs

    :param title: Overall title
    :param pages: Sequence of pages, each string content
    """

    def __init__(self, parent, title="Log", pages=()):
        super(MultiTextViewerDialog, self).__init__(parent)
        self.setWindowTitle(title)
        vbox = QtWidgets.QVBoxLayout()

        self.tabs = QtWidgets.QTabWidget()
        self._browsers = {}
        for heading, content in pages:
            browser = self._text_browser(content)
            self._browsers[heading] = browser
            self.tabs.addTab(browser, heading)

        vbox.addWidget(self.tabs)

        hbox = QtWidgets.QHBoxLayout()
        self.copy_btn = QtWidgets.QPushButton("Copy")
        self.copy_btn.clicked.connect(self._copy)
        hbox.addWidget(self.copy_btn)
        hbox.addStretch(1)
        self.button_box = QtWidgets.QDialogButtonBox(QtWidgets.QDialogButtonBox.Ok)
        self.button_box.accepted.connect(self.close)
        hbox.addWidget(self.button_box)
        vbox.addLayout(hbox)

        self.setLayout(vbox)
        self.resize(700, 500)

    def text(self, heading):
        if heading not in self._browsers:
            raise ValueError("Tab not found: %s" % heading)
        return self._browsers[heading].text()

    def setText(self, heading, content):
        if heading not in self._browsers:
            raise ValueError("Tab not found: %s" % heading)
        scrollbar = self._browsers[heading].verticalScrollBar()
        original_pos = scrollbar.value()
        was_at_end = original_pos == scrollbar.maximum()
        self._browsers[heading].setText(content)
        if was_at_end:
            scrollbar.setValue(scrollbar.maximum())
        else:
            scrollbar.setValue(original_pos)

    def _text_browser(self, content):
        browser = QtWidgets.QTextBrowser()
        browser.setFontFamily("Courier")
        browser.setText(content)
        return browser

    def _copy(self):
        clipboard = QtWidgets.QApplication.clipboard()
        clipboard.setText(self.tabs.currentWidget().toPlainText())

class TextViewerDialog(MultiTextViewerDialog):
    """
    Simple text viewer dialog box
    """

    def __init__(self, parent, title="Log", text=""):
        MultiTextViewerDialog.__init__(self, parent, title, [("", text)])
        self.tabs.tabBar().setVisible(False)

    @property
    def text(self):
        return MultiTextViewerDialog.text(self, "")

    @text.setter
    def text(self, newtext):
        self.setText("", newtext)

class MatrixViewerDialog(QtWidgets.QDialog):
    """
    Dialog box enabling a read-only viewing of a number matrix
    """

    def __init__(self, parent, vals, title="Data", text=""):
        super(MatrixViewerDialog, self).__init__(parent)
        self.setWindowTitle(title)
        vbox = QtWidgets.QVBoxLayout()

        self.table = QtWidgets.QTableWidget(len(vals), len(vals[0]))
        vbox.addWidget(self.table)
        for row, rvals in enumerate(vals):
            for col, val in enumerate(rvals):
                self.table.setItem(row, col, QtWidgets.QTableWidgetItem(str(val)))

        self.text = QtWidgets.QLabel(text)
        vbox.addWidget(self.text)

        self.button_box = QtWidgets.QDialogButtonBox(QtWidgets.QDialogButtonBox.Ok | QtWidgets.QDialogButtonBox.Cancel)
        self.button_box.accepted.connect(self.accept)
        self.button_box.rejected.connect(self.reject)
        vbox.addWidget(self.button_box)

        self.setLayout(vbox)
        self.resize(500, 500)

class GridEditDialog(QtWidgets.QDialog):
    """
    Dialog box enabling a numerical matrix to be edited
    """

    def __init__(self, parent, vals, col_headers=None, row_headers=None, title="Data", text="", expandable=(True, True)):
        super(GridEditDialog, self).__init__(parent)
        self.setWindowTitle(title)
        vbox = QtWidgets.QVBoxLayout()

        from .widgets import NumberGrid # prevent circular import dependency
        self.table = NumberGrid(vals, col_headers=col_headers, row_headers=row_headers, expandable=expandable)
        self.table.sig_changed.connect(self._validate)
        vbox.addWidget(self.table)

        self.text = QtWidgets.QLabel(text)
        vbox.addWidget(self.text)

        self.button_box = QtWidgets.QDialogButtonBox(QtWidgets.QDialogButtonBox.Ok | QtWidgets.QDialogButtonBox.Cancel)
        self.button_box.accepted.connect(self.accept)
        self.button_box.rejected.connect(self.reject)
        self.button_box.button(QtWidgets.QDialogButtonBox.Ok).setEnabled(False)
        vbox.addWidget(self.button_box)

        self.setLayout(vbox)

    def _validate(self):
        self.button_box.button(QtWidgets.QDialogButtonBox.Ok).setEnabled(self.table.valid())

class ChooseFromListDialog(QtWidgets.QDialog):
    """
    Dialog box enabling one item to be chosen from a list
    """

    def __init__(self, parent, values, return_values=None, title="Choose"):
        super(ChooseFromListDialog, self).__init__(parent)
        self.sel_text = None
        self.sel_data = None

        self.setWindowTitle(title)
        vbox = QtWidgets.QVBoxLayout()

        if return_values is None:
            return_values = values
        self._list = QtWidgets.QListWidget(self)
        for value, data in zip(values, return_values):
            item = QtWidgets.QListWidgetItem(value)
            item.setData(QtCore.Qt.UserRole, data)
            self._list.addItem(item)
            
        vbox.addWidget(self._list)
        self._list.itemClicked.connect(self._item_clicked)
        self._list.itemDoubleClicked.connect(self._item_double_clicked)

        self.button_box = QtWidgets.QDialogButtonBox(QtWidgets.QDialogButtonBox.Ok | QtWidgets.QDialogButtonBox.Cancel)
        self.button_box.accepted.connect(self.accept)
        self.button_box.rejected.connect(self.reject)
        self.button_box.button(QtWidgets.QDialogButtonBox.Ok).setEnabled(False)
        vbox.addWidget(self.button_box)

        self.setLayout(vbox)

    def _item_clicked(self, item):
        self.sel_text = item.text()
        self.sel_data = item.data(QtCore.Qt.UserRole)
        self.button_box.button(QtWidgets.QDialogButtonBox.Ok).setEnabled(True)

    def _item_double_clicked(self, item):
        self._item_clicked(item)
        self.accept()

class PlotDialog1d(QtWidgets.QDialog):
    """
    Dialog that shows a 1D plot
    """

    def __init__(self, parent, arr, title="Plot"):
        QtWidgets.QDialog.__init__(self, parent)

        if arr.ndim != 1:
            raise ValueError("Only for 1D plotting")

        self.setWindowTitle(title)
        vbox = QtWidgets.QVBoxLayout()
        self.setLayout(vbox)

        from quantiphyse.gui.plot import Plot
        plot_widget = Plot(self, title)
        plot_widget.add_line(arr)
        vbox.addWidget(plot_widget)




"""
Quantiphyse - Dialog box for user registration

Copyright (c) 2013-2020 University of Oxford

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from __future__ import division, unicode_literals, print_function, absolute_import

import sys
import traceback
import warnings

from PySide2 import QtGui, QtCore, QtWidgets

import quantiphyse.gui.dialogs
from quantiphyse.utils import get_icon, get_version
from quantiphyse._version import __license__

LICENSE_ACCEPTED_KEY = "license_accepted"

def license_accepted():
    settings = QtCore.QSettings()
    if settings.contains(LICENSE_ACCEPTED_KEY):
        try:
            return int(settings.value(LICENSE_ACCEPTED_KEY))
        except:
            # Invalid value - assume not accepted
            return 0
    return 0

def set_license_accepted(accepted=1):
    settings = QtCore.QSettings()
    settings.setValue(LICENSE_ACCEPTED_KEY, accepted)

def check_register():
    """
    Check if the user has accepted the license agreement and if not exit
    the program
    """
    if not license_accepted():
        accepted = LicenseDialog(quantiphyse.gui.dialogs.MAINWIN).exec_()
        set_license_accepted(accepted)

    if not license_accepted():
        # User got the license dialog but did not accept it
        sys.exit(-1)

class LicenseDialog(QtWidgets.QDialog):
    """
    Dialog box which asks a first-time user to accept the license
    """
    def __init__(self, parent=None):
        QtWidgets.QDialog.__init__(self, parent)
        self.setWindowTitle("License Agreement")
        
        layout = QtWidgets.QVBoxLayout()

        hbox = QtWidgets.QHBoxLayout()
        pixmap = QtGui.QPixmap(get_icon("quantiphyse_75.png"))
        lpic = QtWidgets.QLabel(self)
        lpic.setPixmap(pixmap)
        hbox.addWidget(lpic)
        hbox.addStretch(1)
        layout.addLayout(hbox)

        layout.addWidget(QtWidgets.QLabel(""))

        # Welcome
        hbox = QtWidgets.QHBoxLayout()
        hbox.addStretch(1)
        hbox.addWidget(QtWidgets.QLabel("\n<font size=5>Welcome to Quantiphyse %s</font>" % get_version()))
        hbox.addStretch(1)
        layout.addLayout(hbox)

        label = QtWidgets.QLabel("Quantiphyse is <b>free for non-commercial use</b>. If you are interested in using the software "
                             "commercially, please contact the technology transfer company of the University: enquiries@innovation.ox.ac.uk")
        label.setWordWrap(True)
        layout.addWidget(label)

        # License agreement
        edit = QtWidgets.QTextEdit()
        edit.setCurrentFont(QtGui.QFont("Monospace", 8))
        edit.append(__license__)
        edit.append("</pre>")
        edit.moveCursor(QtGui.QTextCursor.Start)
        edit.ensureCursorVisible()
        layout.addWidget(edit)
        
        # Acceptance section
        self.agree_cb = QtWidgets.QCheckBox("I agree to abide by the terms of the Quantiphyse license")
        self.agree_cb.stateChanged.connect(self._agree_changed)
        layout.addWidget(self.agree_cb)

        # Buttons
        self.button_box = QtWidgets.QDialogButtonBox(QtWidgets.QDialogButtonBox.Ok|QtWidgets.QDialogButtonBox.Cancel)
        self.button_box.accepted.connect(self.accept)
        self.button_box.rejected.connect(self.reject)
        self.button_box.button(QtWidgets.QDialogButtonBox.Ok).setEnabled(False)
        layout.addWidget(self.button_box)

        self.setLayout(layout)
        self.setFixedSize(700, 800)

    def _agree_changed(self, state):
        self.button_box.button(QtWidgets.QDialogButtonBox.Ok).setEnabled(state)



"""
Quantiphyse - Tests for supervoxel clustering widget

Copyright (c) 2013-2020 University of Oxford

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import unittest 

import numpy as np

from quantiphyse.test.widget_test import WidgetTest

from .widget import SmoothingWidget

NAME = "test_smooth"
SIGMA = 0.5

class SmoothingWidgetTests(WidgetTest):

    def widget_class(self):
        return SmoothingWidget

    def testNoData(self):
        """ User clicks the generate buttons with no data"""
        self.harmless_click(self.w.run_btn)

    def testNoSmoothing(self):
        """ Check that data is unchanged if sigma = 0 """
        self.ivm.add(self.data_3d, grid=self.grid, name="data_3d")
        self.w.data_combo.setCurrentIndex(0)
        self.processEvents()

        self.w.sigma.spin.setValue(0)
        self.w.output_name.setText(NAME)

        self.harmless_click(self.w.run_btn)
        self.processEvents()

        self.assertTrue(NAME in self.ivm.data)
        self.assertEquals(self.ivm.current_data.name, NAME)
        self.assertFalse(self.error)

        d = self.ivm.data[NAME].raw()
        self.assertTrue(np.all(d == self.data_3d))

    def test3dData(self):
        self.ivm.add(self.data_3d, grid=self.grid, name="data_3d")
        self.w.data_combo.setCurrentIndex(0)
        self.processEvents()

        self.w.sigma.spin.setValue(SIGMA)
        self.w.output_name.setText(NAME)

        self.harmless_click(self.w.run_btn)
        self.processEvents()

        self.assertTrue(NAME in self.ivm.data)
        self.assertEquals(self.ivm.current_data.name, NAME)
        self.assertFalse(self.error)

        d = self.ivm.data[NAME].raw()
        self.assertFalse(np.all(d == self.data_3d))

    def test4dData(self):
        self.ivm.add(self.data_4d, grid=self.grid, name="data_4d")
        self.w.data_combo.setCurrentIndex(0)
        self.processEvents()

        self.w.sigma.spin.setValue(SIGMA)
        self.w.output_name.setText(NAME)

        self.harmless_click(self.w.run_btn)
        self.processEvents()

        self.assertTrue(NAME in self.ivm.data)
        self.assertEquals(self.ivm.current_data.name, NAME)
        self.assertEquals(self.ivm.data[NAME].nvols, self.data_4d.shape[3])
        self.assertFalse(self.error)

        d = self.ivm.data[NAME].raw()
        self.assertFalse(np.all(d == self.data_4d))

if __name__ == '__main__':
    unittest.main()



"""
Quantiphyse - Processes for basic loading/saving of data

Copyright (c) 2013-2020 University of Oxford

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import os

from quantiphyse.utils import QpException
from quantiphyse.data import load, save

from .process import Process

__all__ = ["LoadProcess", "LoadDataProcess", "LoadRoisProcess", "SaveProcess", "SaveAllExceptProcess", "SaveDeleteProcess", "SaveArtifactsProcess"]

class LoadProcess(Process):
    """
    Load data into the IVM
    """
    def __init__(self, ivm, **kwargs):
        Process.__init__(self, ivm, **kwargs)

    def run(self, options):
        rois = options.pop('rois', {})
        data = options.pop('data', {})
        # Force 3D data to be multiple 2D volumes 
        force_mv = options.pop('force-multivol', False)

        for fname, name in list(data.items()) + list(rois.items()):
            qpdata = self._load_file(fname, name)
            if qpdata is not None: 
                if force_mv and qpdata.nvols == 1 and qpdata.grid.shape[2] > 1: 
                    qpdata.set_2dt()
                qpdata.roi = fname in rois
                self.ivm.add(qpdata, make_current=True)

    def _load_file(self, fname, name):
        filepath = self._get_filepath(fname)
        if name is None:
            name = self.ivm.suggest_name(os.path.split(fname)[1].split(".", 1)[0])
        self.debug("  - Loading data '%s' from %s" % (name, filepath))
        try:
            data = load(filepath)
            data.name = name
            return data
        except QpException as exc:
            self.warn("Failed to load data: %s (%s)" % (filepath, str(exc)))

    def _get_filepath(self, fname, folder=None):
        if os.path.isabs(fname):
            return fname
        else:
            if folder is None: folder = self.indir
            return os.path.abspath(os.path.join(folder, fname))

class LoadDataProcess(LoadProcess):
    """
    Process to load data

    Deprecated: use LoadProcess
    """
    def run(self, options):
        LoadProcess.run(self, {'data' : options})
        for key in list(options.keys()): options.pop(key)

class LoadRoisProcess(LoadProcess):
    """
    Process to load ROIs

    Deprecated: use LoadProcess
    """
    def run(self, options):
        LoadProcess.run(self, {'rois' : options})
        for key in list(options.keys()): options.pop(key)

class SaveProcess(Process):
    """
    Save data to file
    """
    def __init__(self, ivm, **kwargs):
        Process.__init__(self, ivm, **kwargs)

    def run(self, options):
        # Note that output-grid is not a valid data name so will not clash
        output_grid = None
        output_grid_name = options.pop("output-grid", None)
        if output_grid_name is not None:
            output_grid_data = self.ivm.data.get(output_grid_name, self.ivm.rois.get(output_grid_name, None))
            if output_grid_data is None:
                raise QpException("No such data found as source of grid: %s" % output_grid_name)
            else:
                output_grid = output_grid_data.grid

        for name in list(options.keys()):
            try:
                fname = options.pop(name, name)
                qpdata = self.ivm.data.get(name, None)
                if qpdata is not None:
                    save(qpdata, fname, grid=output_grid, outdir=self.outdir)
                else:
                    self.warn("Failed to save %s - no such data or ROI found" % name)
            except QpException as exc:
                self.warn("Failed to save %s: %s" % (name, str(exc)))

class SaveAllExceptProcess(Process):
    """
    Save all data to file apart from specified items
    """
    def __init__(self, ivm, **kwargs):
        Process.__init__(self, ivm, **kwargs)

    def run(self, options):
        exceptions = list(options.keys())
        for k in exceptions: options.pop(k)

        for name, qpdata in self.ivm.data.items():
            if name in exceptions: 
                continue
            try:
                save(qpdata, name, outdir=self.outdir)
            except QpException as exc:
                self.warn("Failed to save %s: %s" % (name, str(exc)))
            except:
                import traceback
                traceback.print_exc()

class SaveDeleteProcess(SaveProcess):
    """
    Save data to file and then delete it
    """
    def __init__(self, ivm, **kwargs):
        SaveProcess.__init__(self, ivm, **kwargs)

    def run(self, options):
        options_save = dict(options)
        SaveProcess.run(self, options)

        for name in options_save:
            if name in self.ivm.data: self.ivm.delete(name)

class SaveArtifactsProcess(Process):
    """
    Save 'extras' (previously known as 'artifacts')
    """
    def __init__(self, ivm, **kwargs):
        Process.__init__(self, ivm, **kwargs)

    def run(self, options):
        for name in list(options.keys()):
            fname = options.pop(name)
            if not fname: fname = name
            if name in self.ivm.extras: 
                self.debug("Saving '%s' to %s" % (name, fname))
                self._save_text(str(self.ivm.extras[name]), fname)
            else:
                self.warn("Extra '%s' not found - not saving" % name)

    def _save_text(self, text, fname, ext="txt"):
        if text:
            if "." not in fname: fname = "%s.%s" % (fname, ext)
            if not os.path.isabs(fname):
                fname = os.path.join(self.outdir, fname)
            dirname = os.path.dirname(fname)
            if not os.path.exists(dirname): os.makedirs(dirname)
            with open(fname, "w") as text_file:
                text_file.write(text)



"""
Quantiphyse - Analysis processes for data smoothing

Copyright (c) 2013-2020 University of Oxford

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import numpy as np
import scipy.ndimage.filters

from quantiphyse.processes import Process
from quantiphyse.data import NumpyData

class SmoothingProcess(Process):
    """
    Simple process for Gaussian smoothing
    """
    PROCESS_NAME = "Smooth"

    def __init__(self, ivm, **kwargs):
        Process.__init__(self, ivm, **kwargs)

    def run(self, options):
        data = self.get_data(options)

        output_name = options.pop("output-name", "%s_smoothed" % data.name)
        #kernel = options.pop("kernel", "gaussian")
        order = options.pop("order", 0)
        mode = options.pop("boundary-mode", "reflect")
        sigma = options.pop("sigma", 1.0)

        # Sigma is in mm so scale with data voxel sizes
        if isinstance(sigma, (int, float)):
            sigmas = [float(sigma) / size for size in data.grid.spacing]
        else:
            sigmas = [float(sig) / size for sig, size in zip(sigma, data.grid.spacing)]

        # Smooth multiple volumes independently
        if data.nvols > 1:
            sigmas += [0, ]

        output = self._norm_conv(data.raw(), sigmas, order=order, mode=mode)
        self.ivm.add(NumpyData(output, grid=data.grid, name=output_name), make_current=True)

    def _norm_conv(self, data, sigma, **kwargs):
        """
        Normalized convolution

        This is a way to compensate for data having nan/infinite values.
        Taken from stackoverflow.com/questions/18697532/gaussian-filtering-a-image-with-nan-in-python
        """
        v = data.copy()
        v[~np.isfinite(data)] = 0
        vv = scipy.ndimage.filters.gaussian_filter(v, sigma, **kwargs)

        w = 0*data.copy()+1
        w[~np.isfinite(data)] = 0
        ww = scipy.ndimage.filters.gaussian_filter(w, sigma, **kwargs)

        return vv/ww



"""
Quantiphyse - Miscellaneous generic analysis processes

Copyright (c) 2013-2020 University of Oxford

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import numpy as np
import scipy

from quantiphyse.data import NumpyData

from .process import Process

class RenameProcess(Process):
    """ 
    Rename data  
    """
    
    PROCESS_NAME = "Rename"

    def run(self, options):
        for name in list(options.keys()):
            newname = options.pop(name)
            self.ivm.rename(name, newname)

class DeleteProcess(Process):
    """
    Delete data or ROIs
    """

    PROCESS_NAME = "Delete"

    def __init__(self, ivm, **kwargs):
        Process.__init__(self, ivm, **kwargs)

    def run(self, options):
        for name in list(options.keys()):
            options.pop(name, None)
            if name in self.ivm.data: 
                self.ivm.delete(name)
            else:
                self.warn("Failed to delete %s: No such data or ROI" % name)

class RoiCleanupProcess(Process):
    """
    Fill holes, etc in ROI
    """
    
    PROCESS_NAME = "RoiCleanup"
    
    def __init__(self, ivm, **kwargs):
        Process.__init__(self, ivm, **kwargs)

    def run(self, options):
        roi_name = options.pop('roi', None)
        output_name = options.pop('output-name', "roi-cleaned")
        fill_holes_slice = options.pop('fill-holes-by-slice', None)

        if roi_name is None:
            roi = self.ivm.current_roi
        else:
            roi = self.ivm.rois[roi_name]

        if roi is not None:
            if fill_holes_slice is not None:
                # slice-by-slice hole filling, appropriate when ROIs defined slice-by-slice
                slice_axis = fill_holes_slice
                new = np.copy(roi.raw())
                for slice_idx in range(new.shape[int(slice_axis)]):
                    slices = [slice(None), slice(None), slice(None)]
                    slices[slice_axis] = slice_idx
                    new[slices] = scipy.ndimage.morphology.binary_fill_holes(new[slices])
            
                self.ivm.add(NumpyData(data=new, grid=roi.grid, name=output_name, roi=True))



"""
Quantiphyse - Widgets which perform data clustering

Copyright (c) 2013-2020 University of Oxford

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from __future__ import division, unicode_literals, absolute_import, print_function

import numpy as np
from sklearn.metrics import pairwise

import pyqtgraph as pg
from PySide2 import QtGui, QtCore, QtWidgets

from quantiphyse.data import NumpyData
from quantiphyse.gui.widgets import QpWidget, OverlayCombo, RoiCombo, NumericOption, TitleWidget
from quantiphyse.gui.colors import get_roi_col

from .kmeans import KMeansProcess, MeanValuesProcess

class ClusteringWidget(QpWidget):
    """
    Widget for doing K-means clustering on 3D or 4D data
    """

    def __init__(self, **kwargs):
        super(ClusteringWidget, self).__init__(name="KMeans Clustering", icon="clustering", desc="Generate clusters from 3D or 4D data", group="Clustering", position=2, **kwargs)
        self.curves = {}

    def init_ui(self):
        self.process = KMeansProcess(self.ivm)
        layout = QtWidgets.QVBoxLayout()
        self.setLayout(layout)

        title = TitleWidget(self, help="cluster")
        layout.addWidget(title)
        
        DESC = """
<i>Performs clustering of 3D or 4D data using the K-means algorithm.
PCA reduction is used on 4D data to extract representative curves
"""
        desc = QtWidgets.QLabel(DESC)
        desc.setWordWrap(True)
        layout.addWidget(desc)

        gbox = QtWidgets.QGroupBox()
        gbox.setTitle('Clustering options')

        grid = QtWidgets.QGridLayout()
        gbox.setLayout(grid)

        # Data to cluster
        grid.addWidget(QtWidgets.QLabel("Data"), 0, 0)
        self.data_combo = OverlayCombo(self.ivm)
        self.data_combo.currentIndexChanged.connect(self._data_changed)
        grid.addWidget(self.data_combo, 0, 1)

        grid.addWidget(QtWidgets.QLabel("ROI"), 1, 0)
        self.roi_combo = RoiCombo(self.ivm, none_option=True)
        grid.addWidget(self.roi_combo, 1, 1)

        # Number of clusters inside the ROI
        self.n_clusters = NumericOption("Number of clusters", grid, xpos=2, ypos=0, minval=2, maxval=20, default=4, intonly=True)
        self.n_clusters.spin.setToolTip("")

        # Number of PCA modes
        self.n_pca = NumericOption("Number of PCA modes", grid, xpos=2, ypos=1, minval=1, maxval=10, default=3, intonly=True)
        self.n_pca.spin.setToolTip("")

        # Output ROI name
        grid.addWidget(QtWidgets.QLabel("Output name"), 2, 0)
        self.output_name = QtWidgets.QLineEdit("clusters")
        grid.addWidget(self.output_name, 2, 1)
        layout.addWidget(gbox)

        # Run clustering button
        hbox = QtWidgets.QHBoxLayout()
        self.run_btn = QtWidgets.QPushButton('Run', self)
        self.run_btn.clicked.connect(self.run_clustering)
        hbox.addWidget(self.run_btn)
        hbox.addStretch(1)
        layout.addLayout(hbox)

        # Plot window, showing representative curves for 4D data
        self.show_curves_btn = QtWidgets.QPushButton('Show representative curves', self)
        self.show_curves_btn.clicked.connect(self._show_curves)
        layout.addWidget(self.show_curves_btn)
        self.plotwin = pg.GraphicsLayoutWidget()
        self.plotwin.setBackground(background=None)
        self.plot = self.plotwin.addPlot(title="Cluster representative curves")
        self.plot.setLabel('left', "Signal Enhancement")
        self.plotwin.setVisible(False)        
        layout.addWidget(self.plotwin)

        # Statistics
        self.show_count_btn = QtWidgets.QPushButton('Show voxel counts', self)
        self.show_count_btn.clicked.connect(self._show_counts)
        layout.addWidget(self.show_count_btn)
        self.stats_gbox = QtWidgets.QGroupBox()
        self.stats_gbox.setTitle('Voxel count')

        self.count_table = QtGui.QStandardItemModel()
        self.count_view = QtWidgets.QTableView()
        self.count_view.resizeColumnsToContents()
        self.count_view.setModel(self.count_table)

        hbox = QtWidgets.QHBoxLayout()
        hbox.addWidget(self.count_view)
        self.stats_gbox.setLayout(hbox)
        self.stats_gbox.setVisible(False)
        layout.addWidget(self.stats_gbox)

        # Merge regions
        self.show_merge_btn = QtWidgets.QPushButton('Show merge options', self)
        self.show_merge_btn.clicked.connect(self._show_merge)
        layout.addWidget(self.show_merge_btn)
        
        self.merge_gbox = QtWidgets.QGroupBox()
        self.merge_gbox.setTitle('Merge regions')
        vbox = QtWidgets.QVBoxLayout()
        self.merge_gbox.setLayout(vbox)

        hbox = QtWidgets.QHBoxLayout()
        self.merge_btn = QtWidgets.QPushButton('Merge', self)
        self.merge_btn.clicked.connect(self._run_merge)
        hbox.addWidget(self.merge_btn)
        hbox.addWidget(QtWidgets.QLabel('Merge region '))
        self.merge_region1 = QtWidgets.QLineEdit('1', self)
        hbox.addWidget(self.merge_region1)
        hbox.addWidget(QtWidgets.QLabel(' with '))
        self.merge_region2 = QtWidgets.QLineEdit('2', self)
        hbox.addWidget(self.merge_region2)
        vbox.addLayout(hbox)

        hbox = QtWidgets.QHBoxLayout()
        self.auto_merge_btn = QtWidgets.QPushButton('AutoMerge', self)
        self.auto_merge_btn.clicked.connect(self._run_automerge)
        hbox.addWidget(self.auto_merge_btn)
        hbox.addStretch(1)
        vbox.addLayout(hbox)

        self.merge_gbox.setVisible(False)
        layout.addWidget(self.merge_gbox)

        layout.addStretch(1)

    def activate(self):
        self.ivl.sig_focus_changed.connect(self._focus_changed)
        self.ivm.sig_current_roi.connect(self._current_roi_changed)
        self.ivm.sig_main_data.connect(self._main_data_changed)
        self._data_changed()

    def deactivate(self):
        self.ivl.sig_focus_changed.disconnect(self._focus_changed)
        self.ivm.sig_current_roi.disconnect(self._current_roi_changed)
        self.ivm.sig_main_data.disconnect(self._main_data_changed)

    def _current_roi_changed(self, roi):
        if roi is not None and roi.name != self.output_name.text():
            self.roi_combo.setCurrentIndex(self.roi_combo.findText(roi.name))
        
    def _main_data_changed(self, data):
        if data is not None:
            idx = self.data_combo.findText(data.name)
        else:
            idx = 0
        self.data_combo.setCurrentIndex(idx)

    def _data_changed(self):
        data = self.ivm.data.get(self.data_combo.currentText(), None)
        if data is not None:
            is4d = data.nvols > 1
            self.debug("Number of vols: %i, 4d=%s", data.nvols, is4d)
            self.n_pca.label.setVisible(is4d)
            self.n_pca.spin.setVisible(is4d)
            self.show_curves_btn.setVisible(is4d)
            self.plotwin.setVisible(is4d and self.plotwin.isVisible())
            self.auto_merge_btn.setEnabled(is4d)
        self.run_btn.setEnabled(data is not None)

    def _focus_changed(self):
        self._update_voxel_count()

    def _show_curves(self):
        if self.plotwin.isVisible():
            self.plotwin.setVisible(False)
            self.show_curves_btn.setText("Show representative curves")
        else:
            self.plotwin.setVisible(True)
            self.show_curves_btn.setText("Hide representative curves")
            
    def _show_counts(self):
        if self.stats_gbox.isVisible():
            self.stats_gbox.setVisible(False)
            self.show_count_btn.setText("Show voxel counts")
        else:
            self.stats_gbox.setVisible(True)
            self.show_count_btn.setText("Hide voxel counts")
        
    def _show_merge(self):
        if self.merge_gbox.isVisible():
            self.merge_gbox.setVisible(False)
            self.show_merge_btn.setText("Show merge_options")
        else:
            self.merge_gbox.setVisible(True)
            self.show_merge_btn.setText("Hide merge options")
        
    def batch_options(self):
        options = {
            "data" : self.data_combo.currentText(),
            "roi" :  self.roi_combo.currentText(),
            "n-clusters" : self.n_clusters.spin.value(),
            "output-name" : self.output_name.text(),
            "invert-roi" : False,
        }

        if options["roi"] == "<none>":
            del options["roi"]

        if self.n_pca.label.isVisible():
            # 4D PCA options
            options["n-pca"] = self.n_pca.spin.value()
            options["reduction"] = "pca"
            options["norm-data"] = True

        return "KMeans", options

    def run_clustering(self):
        """
        Run kmeans clustering using normalised PCA modes
        """
        options = self.batch_options()[1]

        self.process.run(options)
        self._update_voxel_count()
        if self.n_pca.label.isVisible():
            self.update_plot()     

    def _update_voxel_count(self):
        self.count_table.clear()
        self.count_table.setVerticalHeaderItem(0, QtGui.QStandardItem("Voxel count"))

        roi = self.ivm.rois.get(self.output_name.text(), None)
        if roi is not None:
            col_idx = 0
            for region, name in roi.regions.items():
                self.count_table.setHorizontalHeaderItem(col_idx, QtGui.QStandardItem(name))

                # Volume count
                voxel_count = np.sum(roi.raw() == region)
                self.count_table.setItem(0, col_idx, QtGui.QStandardItem(str(np.around(voxel_count))))
                col_idx += 1

    def update_plot(self):
        """
        Plot the cluster curves
        :return:
        """
        # Clear graph
        self.plot.clear()
        self.plot.setLabel('bottom', "Volume", "")
        if self.plot.legend is not None:
            # Work around pyqtgraph legend bug - the legend is recreated multiple times!
            # https://stackoverflow.com/questions/42792858/pyqtgraph-delete-persisting-legend-in-pyqt4-gui
            self.plot.legend.scene().removeItem(self.plot.legend)
        data = self.ivm.data.get(self.data_combo.currentText(), None)
        roi = self.ivm.rois.get(self.output_name.text(), None)
        if roi is not None and data is not None and data.nvols > 1:
            # Generate mean curve for each cluster
            self._generate_cluster_means(roi, data)

            self.plot.addLegend()

            # Plotting using single or multiple plots
            for region, name in roi.regions.items():
                if np.sum(self.curves[region]) == 0:
                    continue

                pencol = get_roi_col(roi, region)
                curve = self.plot.plot(pen=pencol, width=8.0, name=name)
                xx = np.arange(len(self.curves[region]))
                curve.setData(xx, self.curves[region])

    def _run_merge(self):
        m1 = int(self.merge_region1.text())
        m2 = int(self.merge_region2.text())
        self._merge(m1, m2)

    def _run_automerge(self):
        # Use PCA features or true curves?
        # Mean features from each cluster
        # Distance matrix between features
        if len(self.curves) < 2: return

        curvemat = np.zeros((len(self.curves), self.ivm.main.nvols))
        row_region = {}
        idx = 0
        for region, curve in self.curves.items():
            row_region[idx] = region
            curvemat[idx, :] = curve
            idx += 1
        distmat = pairwise.euclidean_distances(curvemat)
        distmat[distmat == 0] = np.inf
        loc1 = np.where(distmat == distmat.min())[0]
        self._merge(row_region[loc1[0]], row_region[loc1[1]])

    def _generate_cluster_means(self, roi, data):
        """
        Generate the mean curves for each cluster
        Returns:
        """
        self.curves = {}
        for region in roi.regions:
            roi_data = data.mask(roi, region=region, output_flat=True)
            mean = np.median(roi_data, axis=0)
            self.curves[region] = mean

    def _merge(self, m1, m2):
        roi = self.ivm.rois.get(self.output_name.text(), None)
        if roi is not None:
            roi_data = roi.raw()
            roi_data[roi_data == m1] = m2

            # signal the change
            self.ivm.add(NumpyData(roi_data, grid=roi.grid, name=self.output_name.text(), roi=True), make_current=True)

            # replot
            self.update_plot()
            self._update_voxel_count()

class MeanValuesWidget(QpWidget):
    """
    Convert a data + multi-level ROI into mean values data set
    """
    def __init__(self, **kwargs):
        super(MeanValuesWidget, self).__init__(name="Mean in ROI", icon="meanvals", 
                                               desc="Replace data with its mean value within each ROI region", 
                                               group="ROIs", **kwargs)

        layout = QtWidgets.QVBoxLayout()

        title = TitleWidget(self, "Generate Mean Values Data", help="mean_values")
        layout.addWidget(title)
        
        desc = QtWidgets.QLabel("This widget will convert the current data set into a "
                            "new data set in which each ROI region contains the mean "
                            "value for that region.\n\nThis is generally only useful for "
                            "multi-level ROIs such as clusters or supervoxels")
        desc.setWordWrap(True)
        layout.addWidget(desc)

        hbox = QtWidgets.QHBoxLayout()
        gbox = QtWidgets.QGroupBox()
        gbox.setTitle("Options")
        grid = QtWidgets.QGridLayout()
        gbox.setLayout(grid)
        
        grid.addWidget(QtWidgets.QLabel("Data"), 0, 0)
        self.ovl = OverlayCombo(self.ivm)
        self.ovl.currentIndexChanged.connect(self._data_changed)
        grid.addWidget(self.ovl, 0, 1)
        grid.addWidget(QtWidgets.QLabel("ROI regions"), 1, 0)
        self.roi = RoiCombo(self.ivm)
        grid.addWidget(self.roi, 1, 1)
        grid.addWidget(QtWidgets.QLabel("Output name"), 2, 0)
        self.output_name = QtWidgets.QLineEdit()
        grid.addWidget(self.output_name, 2, 1)

        btn = QtWidgets.QPushButton('Generate', self)
        btn.clicked.connect(self._generate)
        grid.addWidget(btn, 2, 0)
        hbox.addWidget(gbox)
        hbox.addStretch(1)
        layout.addLayout(hbox)
        layout.addStretch(1)
        self.setLayout(layout)

    def _data_changed(self):
        name = self.ovl.currentText()
        if name:
            self.output_name.setText(name + "_means")

    def batch_options(self):
        options = {
            "roi" : self.roi.currentText(),
            "data" : self.ovl.currentText(),
            "output-name" :  self.output_name.text()
        }
        return "MeanValues", options

    def _generate(self):
        options = self.batch_options()[1]

        if not options["data"]:
            QtWidgets.QMessageBox.warning(self, "No data selected", "Load data to generate mean values from", QtWidgets.QMessageBox.Close)
            return
        if not options["roi"]:
            QtWidgets.QMessageBox.warning(self, "No ROI selected", "Load an ROI for mean value regions", QtWidgets.QMessageBox.Close)
            return
        
        process = MeanValuesProcess(self.ivm)
        process.run(options)



"""
Quantiphyse - Tests for OrthoSlice class

Copyright (c) 2013-2020 University of Oxford

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import unittest

import numpy as np

from quantiphyse.data import DataGrid, NumpyData, OrthoSlice

GRIDSIZE = 5
SLICEPOS = 2
XAXIS, YAXIS, ZAXIS = 0, 1, 2

class OrthoSliceTest(unittest.TestCase):

    def testOrthoY(self):
        grid = DataGrid((GRIDSIZE, GRIDSIZE, GRIDSIZE), np.identity(4))
        YD, XD, ZD = np.meshgrid(range(GRIDSIZE), range(GRIDSIZE), range(GRIDSIZE))

        plane = OrthoSlice(grid, YAXIS, SLICEPOS)
        self.assertEquals(tuple(plane.origin), (0, SLICEPOS, 0))
        self.assertEquals(len(plane.basis), 2)
        self.assertTrue((1, 0, 0) in plane.basis)
        self.assertTrue((0, 0, 1) in plane.basis)
        
    def testOrthoX(self):
        grid = DataGrid((GRIDSIZE, GRIDSIZE, GRIDSIZE), np.identity(4))
        plane = OrthoSlice(grid, XAXIS, SLICEPOS)
        self.assertEquals(tuple(plane.origin), (SLICEPOS, 0, 0))
        self.assertEquals(len(plane.basis), 2)
        self.assertTrue((0, 1, 0) in plane.basis)
        self.assertTrue((0, 0, 1) in plane.basis)
        
    def testOrthoZ(self):
        grid = DataGrid((GRIDSIZE, GRIDSIZE, GRIDSIZE), np.identity(4))
        plane = OrthoSlice(grid, ZAXIS, SLICEPOS)
        self.assertEquals(tuple(plane.origin), (0, 0, SLICEPOS))
        self.assertEquals(len(plane.basis), 2)
        self.assertTrue((0, 1, 0) in plane.basis)
        self.assertTrue((1, 0, 0) in plane.basis)
        
    def testGenericX(self):
        trans = np.array([
            [0.3, 0.2, 1.7, 0],
            [0.1, 2.1, 0.11, 0],
            [2.2, 0.7, 0.3, 0],
            [0, 0, 0, 1]
        ])

        grid = DataGrid((GRIDSIZE, GRIDSIZE, GRIDSIZE), trans)
        origin = list(SLICEPOS * trans[:3,0])
        plane = OrthoSlice(grid, XAXIS, SLICEPOS)
        self.assertAlmostEquals(list(plane.origin), origin)
        self.assertEquals(len(plane.basis), 2)
        self.assertTrue(tuple(trans[:3, 2]) in plane.basis)
        self.assertTrue(tuple(trans[:3, 1]) in plane.basis)

    def testGenericY(self):
        trans = np.array([
            [0.3, 0.2, 1.7, 0],
            [0.1, 2.1, 0.11, 0],
            [2.2, 0.7, 0.3, 0],
            [0, 0, 0, 1]
        ])

        grid = DataGrid((GRIDSIZE, GRIDSIZE, GRIDSIZE), trans)
        origin = list(SLICEPOS * trans[:3,1])
        plane = OrthoSlice(grid, YAXIS, SLICEPOS)
        self.assertAlmostEquals(list(plane.origin), origin)
        self.assertEquals(len(plane.basis), 2)
        self.assertTrue(tuple(trans[:3, 0]) in plane.basis)
        self.assertTrue(tuple(trans[:3, 2]) in plane.basis)

    def testGenericZ(self):
        trans = np.array([
            [0.3, 0.2, 1.7, 0],
            [0.1, 2.1, 0.11, 0],
            [2.2, 0.7, 0.3, 0],
            [0, 0, 0, 1]
        ])

        grid = DataGrid((GRIDSIZE, GRIDSIZE, GRIDSIZE), trans)
        origin = list(SLICEPOS * trans[:3,2])
        plane = OrthoSlice(grid, ZAXIS, SLICEPOS)
        self.assertAlmostEquals(list(plane.origin), origin)
        self.assertEquals(len(plane.basis), 2)
        self.assertTrue(tuple(trans[:3, 0]) in plane.basis)
        self.assertTrue(tuple(trans[:3, 1]) in plane.basis)
    """
    def testSliceIdenticalZ(self):
        grid = DataGrid((GRIDSIZE, GRIDSIZE, GRIDSIZE), np.identity(4))
        plane = OrthoSlice(grid, ZAXIS, SLICEPOS)
        data = np.random.rand(*grid.shape)
        plane.slice_data(data, grid)

    def testSliceIdenticalY(self):
        grid = DataGrid((GRIDSIZE, GRIDSIZE, GRIDSIZE), np.identity(4))
        plane = OrthoSlice(grid, YAXIS, SLICEPOS)
        data = np.random.rand(*grid.shape)
        plane.slice_data(data, grid)
    """
    def testHighRes(self):
        grid = DataGrid((GRIDSIZE, GRIDSIZE, GRIDSIZE), np.identity(4))
        plane = OrthoSlice(grid, YAXIS, SLICEPOS)
        data = np.random.rand(GRIDSIZE*2, GRIDSIZE*2, GRIDSIZE*2)
        datagrid = DataGrid((GRIDSIZE*2, GRIDSIZE*2, GRIDSIZE*2), np.identity(4)/2)
        qpd = NumpyData(data, name="test", grid=datagrid)
        qpd.slice_data(plane)

    def testOrtho(self):
        grid = DataGrid((GRIDSIZE, GRIDSIZE, GRIDSIZE), np.identity(4))
        YD, XD, ZD = np.meshgrid(range(GRIDSIZE), range(GRIDSIZE), range(GRIDSIZE))
        
        plane = OrthoSlice(grid, YAXIS, SLICEPOS)
        xdata, _, _, _ = NumpyData(XD, name="test", grid=grid).slice_data(plane)
        ydata, _, _, _ = NumpyData(YD, name="test", grid=grid).slice_data(plane)
        zdata, _, _, _ = NumpyData(ZD, name="test", grid=grid).slice_data(plane)

        self.assertTrue(np.all(ydata == SLICEPOS))
        for x in range(GRIDSIZE):
           self.assertTrue(np.all(xdata[x,:] == x))
           self.assertTrue(np.all(zdata[:,x] == x))

    def testOrthoSwapAxis(self):
        grid = DataGrid((GRIDSIZE, GRIDSIZE, GRIDSIZE), np.identity(4))
        plane = OrthoSlice(grid, YAXIS, SLICEPOS)

        # Swap Y and Z axes
        affine = np.array([
            [1, 0, 0, 0],
            [0, 0, 1, 0],
            [0, 1, 0, 0],
            [0, 0, 0, 1]
        ])
        datagrid = DataGrid((GRIDSIZE, GRIDSIZE, GRIDSIZE), affine)
        YD, XD, ZD = np.meshgrid(range(GRIDSIZE), range(GRIDSIZE), range(GRIDSIZE))

        xdata, _, _, _ = NumpyData(XD, name="test", grid=datagrid).slice_data(plane)
        ydata, _, _, _ = NumpyData(YD, name="test", grid=datagrid).slice_data(plane)
        zdata, _, _, _ = NumpyData(ZD, name="test", grid=datagrid).slice_data(plane)

        self.assertTrue(np.all(zdata == SLICEPOS))
        for x in range(GRIDSIZE):
            self.assertTrue(np.all(xdata[x,:] == x))
            self.assertTrue(np.all(ydata[:,x] == x))

    def testOrthoReversed(self):
        grid = DataGrid((GRIDSIZE, GRIDSIZE, GRIDSIZE), np.identity(4))
        plane = OrthoSlice(grid, YAXIS, SLICEPOS)

        # Invert Z axis
        affine = np.array([
            [1, 0, 0, 0],
            [0, 1, 0, 0],
            [0, 0, -1, GRIDSIZE-1],
            [0, 0, 0, 1]
        ])
        datagrid = DataGrid((GRIDSIZE, GRIDSIZE, GRIDSIZE), affine)
        YD, XD, ZD = np.meshgrid(range(GRIDSIZE), range(GRIDSIZE), range(GRIDSIZE))

        xdata, _, _, _ = NumpyData(XD, name="test", grid=datagrid).slice_data(plane)
        ydata, _, _, _ = NumpyData(YD, name="test", grid=datagrid).slice_data(plane)
        zdata, _, transv, offset = NumpyData(ZD, name="test", grid=datagrid).slice_data(plane)
        
        # Reversal is reflected in the transformation
        self.assertTrue(np.all(transv == [[1, 0], [0, -1]]))

        self.assertTrue(np.all(ydata == SLICEPOS))
        for x in range(GRIDSIZE):
            self.assertTrue(np.all(xdata[x,:] == x))
            self.assertTrue(np.all(zdata[:,x] == x))

    def testOrthoOffset(self):
        grid = DataGrid((GRIDSIZE, GRIDSIZE, GRIDSIZE), np.identity(4))
        plane = OrthoSlice(grid, YAXIS, SLICEPOS)

        # Offset X axis
        affine = np.array([
            [1, 0, 0, 2],
            [0, 1, 0, 0],
            [0, 0, 1, 0],
            [0, 0, 0, 1]
        ])
        datagrid = DataGrid((GRIDSIZE, GRIDSIZE, GRIDSIZE), affine)
        YD, XD, ZD = np.meshgrid(range(GRIDSIZE), range(GRIDSIZE), range(GRIDSIZE))

        xdata, _, _, _ = NumpyData(XD, name="test", grid=datagrid).slice_data(plane)
        ydata, _, _, _ = NumpyData(YD, name="test", grid=datagrid).slice_data(plane)
        zdata, _, transv, offset = NumpyData(ZD, name="test", grid=datagrid).slice_data(plane)

        self.assertTrue(np.all(ydata == SLICEPOS))
        for x in range(GRIDSIZE):
            self.assertTrue(np.all(xdata[x,:] == x))
            self.assertTrue(np.all(zdata[:,x] == x))

if __name__ == '__main__':
    unittest.main()
